from pathlib import Path
import argparse

import json
from PIL import Image

from .builtin_meta import COCO_CATEGORIES


class Yolo2Coco:

    def __init__(self, root, split):
        self.info = {
            'year': 2021,
            'version': '1.0',
            'description': 'For object detection',
            'date_created': '2021',
        }
        self.licenses = [{
            'id': 1,
            'name': 'GNU General Public License v3.0',
            'url': 'https://github.com/zhiqwang/yolov5-rt-stack/blob/master/LICENSE',
        }]
        self.type = 'instances'
        self.split = split
        self.root_path = Path(root)
        self.label_path = self.root_path.joinpath('labels')
        self.annotation_root = self.root_path.joinpath('annotations')
        Path(self.annotation_root).mkdir(parents=True, exist_ok=True)

        self.categories = [{
            'id': coco_category['id'],
            'name': coco_category['name'],
            'supercategory': coco_category['supercategory'],
        } for coco_category in COCO_CATEGORIES]

    def generate(self, coco_type='instances', annotation_format='bbox'):
        label_paths = sorted(self.label_path.rglob('*.txt'))
        images, annotations = self._get_image_annotation_pairs(
            label_paths,
            annotation_format=annotation_format,
        )
        json_data = {
            'info': self.info,
            'images': images,
            'licenses': self.licenses,
            'type': self.type,
            'annotations': annotations,
            'categories': self.categories,
        }
        output_path = self.annotation_root.joinpath(f'{coco_type}_{self.split}.json')
        with open(output_path, 'w') as json_file:
            json.dump(json_data, json_file, sort_keys=True)

    def _get_image_annotation_pairs(self, label_paths, annotation_format='bbox'):
        images = []
        annotations = []
        annotation_id = 0
        for img_id, label_path in enumerate(label_paths, 1):
            img_path = str(label_path).replace('labels', 'images').replace('txt', 'jpg')
            img = Image.open(img_path)
            width, height = img.size

            images.append({
                'date_captured': '2021',
                'file_name': str(Path(img_path).relative_to(self.root_path)),
                'id': img_id,
                'license': 1,
                'url': '',
                'height': height,
                'width': width,
            })

            with open(label_path, 'r') as f:
                for line in f:
                    label_info = line.strip().split()
                    assert len(label_info) == 5
                    annotation_id += 1

                    category_id, vertex_info = label_info[0], label_info[1:]
                    category_id = self.categories[int(category_id)]['id']
                    if annotation_format == 'bbox':
                        segmentation, bbox, area = self._get_annotation(vertex_info, height, width)
                    else:
                        raise NotImplementedError

                    annotations.append({
                        'segmentation': segmentation,
                        'area': area,
                        'iscrowd': 0,
                        'image_id': img_id,
                        'bbox': bbox,
                        'category_id': category_id,
                        'id': annotation_id,
                    })

        return images, annotations

    @staticmethod
    def _get_annotation(vertex_info, height, width):

        cx, cy, w, h = [float(i) for i in vertex_info]
        cx = cx * width
        cy = cy * height
        w = w * width
        h = h * height
        x = cx - w / 2
        y = cy - h / 2

        segmentation = [[x, y, x + w, y, x + w, y + h, x, y + h]]
        area = w * h

        bbox = [x, y, w, h]
        return segmentation, bbox, area


if __name__ == "__main__":
    parser = argparse.ArgumentParser('Datasets converter from yolo to coco', add_help=False)

    parser.add_argument('--data_path', default='../coco128',
                        help='Dataset root path')
    parser.add_argument('--split', default='train2017',
                        help='Dataset split part, optional: [train2017, val2017]')

    args = parser.parse_args()

    converter = Yolo2Coco(args.data_path, args.split)
    converter.generate()



import unittest
import torch
import pytorch_lightning as pl

from yolort.models import YOLOLitWrapper
from yolort.models.yolo import yolov5_darknet_pan_s_r31
from yolort.models.transform import nested_tensor_from_tensor_list
from yolort.datasets import DetectionDataModule

from .torch_utils import image_preprocess
from .dataset_utils import DummyCOCODetectionDataset

from typing import Dict


class EngineTester(unittest.TestCase):
    def test_train(self):
        # Read Image using TorchVision.io Here
        # Do forward over image
        img_name = "test/assets/zidane.jpg"
        img_tensor = image_preprocess(img_name)
        self.assertEqual(img_tensor.ndim, 3)
        # Add a dummy image to train
        img_dummy = torch.rand((3, 416, 360), dtype=torch.float32)

        images = nested_tensor_from_tensor_list([img_tensor, img_dummy])
        targets = torch.tensor([[0, 7, 0.3790, 0.5487, 0.3220, 0.2047],
                                [0, 2, 0.2680, 0.5386, 0.2200, 0.1779],
                                [0, 3, 0.1720, 0.5403, 0.1960, 0.1409],
                                [0, 4, 0.2240, 0.4547, 0.1520, 0.0705]], dtype=torch.float)

        model = yolov5_darknet_pan_s_r31(num_classes=12)
        model.train()
        out = model(images, targets)
        self.assertIsInstance(out, Dict)
        self.assertIsInstance(out["cls_logits"], torch.Tensor)
        self.assertIsInstance(out["bbox_regression"], torch.Tensor)
        self.assertIsInstance(out["objectness"], torch.Tensor)

    def test_train_one_step(self):
        # Load model
        model = YOLOLitWrapper()
        model.train()

        # Setup the DataModule
        train_dataset = DummyCOCODetectionDataset(num_samples=128)
        datamodule = DetectionDataModule(train_dataset, batch_size=16)
        # Trainer
        trainer = pl.Trainer(max_epochs=1)
        trainer.fit(model, datamodule)

    def test_inference(self):
        # Infer over an image
        img_name = "test/assets/zidane.jpg"
        img_input = image_preprocess(img_name)
        self.assertEqual(img_input.ndim, 3)

        model = YOLOLitWrapper(pretrained=True)
        model.eval()

        out = model([img_input])
        self.assertIsInstance(out, list)
        self.assertIsInstance(out[0], Dict)
        self.assertIsInstance(out[0]["boxes"], torch.Tensor)
        self.assertIsInstance(out[0]["labels"], torch.Tensor)
        self.assertIsInstance(out[0]["scores"], torch.Tensor)


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2021, Zhiqiang Wang. All Rights Reserved.
from torch import nn

from .common import Conv
from .pl_wrapper import YOLOLitWrapper

from ..utils.activations import Hardswish


def yolov5s(**kwargs):
    model = YOLOLitWrapper(arch="yolov5_darknet_pan_s_r31", **kwargs)
    return model


def yolov5m(**kwargs):
    model = YOLOLitWrapper(arch="yolov5_darknet_pan_m_r31", **kwargs)
    return model


def yolov5l(**kwargs):
    model = YOLOLitWrapper(arch="yolov5_darknet_pan_l_r31", **kwargs)
    return model


def yolov5_onnx(pretrained=False, progress=True, num_classes=80, **kwargs):

    model = yolov5s(pretrained=pretrained, progress=progress, num_classes=num_classes, **kwargs)
    for m in model.modules():
        m._non_persistent_buffers_set = set()  # pytorch 1.6.0 compatibility
        if isinstance(m, Conv) and isinstance(m.act, nn.Hardswish):
            m.act = Hardswish()  # assign activation

    return model



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.
# Modified by Zhiqiang Wang (zhiqwang@foxmail.com)
import argparse
from pathlib import Path

import pytorch_lightning as pl

from .datasets import VOCDetectionDataModule
from . import models


def get_args_parser():
    parser = argparse.ArgumentParser('You only look once detector', add_help=False)

    parser.add_argument('--arch', default='yolov5s',
                        help='model structure to train')
    parser.add_argument('--data_path', default='./data-bin',
                        help='dataset')
    parser.add_argument('--dataset_type', default='coco',
                        help='dataset')
    parser.add_argument('--dataset_mode', default='instances',
                        help='dataset mode')
    parser.add_argument('--years', default=['2017'], nargs='+',
                        help='dataset year')
    parser.add_argument('--train_set', default='train',
                        help='set of train')
    parser.add_argument('--val_set', default='val',
                        help='set of val')
    parser.add_argument('--batch_size', default=32, type=int,
                        help='images per gpu, the total batch size is $NGPU x batch_size')
    parser.add_argument('--max_epochs', default=1, type=int, metavar='N',
                        help='number of total epochs to run')
    parser.add_argument('--num_gpus', default=1, type=int, metavar='N',
                        help='number of gpu utilizing (default: 1)')
    parser.add_argument('--num_workers', default=4, type=int, metavar='N',
                        help='number of data loading workers (default: 4)')
    parser.add_argument('--print_freq', default=20, type=int,
                        help='print frequency')
    parser.add_argument('--output_dir', default='.',
                        help='path where to save')
    return parser


def main(args):
    # Load the data
    datamodule = VOCDetectionDataModule.from_argparse_args(args)

    # Build the model
    model = models.__dict__[args.arch](num_classes=datamodule.num_classes)

    # Create the trainer. Run twice on data
    trainer = pl.Trainer(max_epochs=args.max_epochs, gpus=args.num_gpus)

    # Train the model
    trainer.fit(model, datamodule=datamodule)


if __name__ == "__main__":
    parser = argparse.ArgumentParser('YOLOv5 training and evaluation script', parents=[get_args_parser()])
    args = parser.parse_args()
    if args.output_dir:
        Path(args.output_dir).mkdir(parents=True, exist_ok=True)

    main(args)



# Modified from ultralytics/yolov5 by Zhiqiang Wang
import torch
from torch import nn, Tensor

from torchvision.ops import batched_nms

from . import _utils as det_utils

from typing import Tuple, List, Dict


class YoloHead(nn.Module):
    def __init__(self, in_channels: List[int], num_anchors: int, num_classes: int):
        super().__init__()
        self.num_anchors = num_anchors  # anchors
        self.num_outputs = num_classes + 5  # number of outputs per anchor

        self.head = nn.ModuleList(
            nn.Conv2d(ch, self.num_outputs * self.num_anchors, 1) for ch in in_channels)  # output conv

    def get_result_from_head(self, features: Tensor, idx: int) -> Tensor:
        """
        This is equivalent to self.head[idx](features),
        but torchscript doesn't support this yet
        """
        num_blocks = 0
        for m in self.head:
            num_blocks += 1
        if idx < 0:
            idx += num_blocks
        i = 0
        out = features
        for module in self.head:
            if i == idx:
                out = module(features)
            i += 1
        return out

    def forward(self, x: List[Tensor]) -> List[Tensor]:
        all_pred_logits: List[Tensor] = []  # inference output

        for i, features in enumerate(x):
            pred_logits = self.get_result_from_head(features, i)

            # Permute output from (N, A * K, H, W) to (N, A, H, W, K)
            N, _, H, W = pred_logits.shape
            pred_logits = pred_logits.view(N, self.num_anchors, -1, H, W)
            pred_logits = pred_logits.permute(0, 1, 3, 4, 2)  # Size=(N, A, H, W, K)

            all_pred_logits.append(pred_logits)

        return all_pred_logits


class SetCriterion(nn.Module):
    """This class computes the loss for YOLOv5.
    Arguments:
        variances:
    """
    __annotations__ = {
        'box_coder': det_utils.BoxCoder,
    }

    def __init__(
        self,
        strides: List[int],
        anchor_grids: List[List[int]],
        box: float = 0.05,  # box loss gain
        cls: float = 0.5,  # cls loss gain
        cls_pw: float = 1.0,  # cls BCELoss positive_weight
        obj: float = 1.0,  # obj loss gain (scale with pixels)
        obj_pw: float = 1.0,  # obj BCELoss positive_weight
        anchor_threshold: float = 4.0,  # anchor-multiple threshold
        iou_ratio: float = 1.0,  # iou loss ratio (obj_loss = 1.0 or iou)
        fl_gamma: float = 0.0,  # focal loss gamma
        layer_balance: List[float] = [4.0, 1.0, 0.4],
    ) -> None:
        """
        Arguments:
            weights (4-element tuple)
            fg_iou_thresh (float)
            bg_iou_thresh (float)
            allow_low_quality_matches (bool)
        """
        super().__init__()
        self.strides = strides
        self.anchor_grids = anchor_grids
        self.anchor_threshold = anchor_threshold
        self.fl_gamma = fl_gamma
        self.layer_balance = layer_balance

        self.cls_pw = cls_pw
        self.obj_pw = obj_pw
        self.cls = cls
        self.obj = obj
        self.box = box

        self.iou_ratio = iou_ratio

    def forward(
        self,
        head_outputs: List[Tensor],
        targets: Tensor,
    ) -> Dict[str, Tensor]:
        """ This performs the loss computation.
        Parameters:
            head_outputs: dict of tensors, see the output specification of the model for the format
            targets: list of dicts, such that len(targets) == batch_size.
                    The expected keys in each dict depends on the losses applied, see each loss' doc
        """
        targets_cls, targets_box, indices, anchors = self.select_training_samples(head_outputs, targets)
        losses = self.compute_loss(head_outputs, targets_cls, targets_box, indices, anchors)

        return losses

    def select_training_samples(
        self,
        head_outputs: List[Tensor],
        targets: Tensor,
    ) -> Tuple[List[Tensor], List[Tensor], List[Tuple[Tensor, Tensor, Tensor, Tensor]], List[Tensor]]:
        # get boxes indices for each anchors
        device = head_outputs[0].device

        num_layers = len(head_outputs)
        anchors = torch.as_tensor(self.anchor_grids, dtype=torch.float32, device=device)
        strides = torch.as_tensor(self.strides, dtype=torch.float32, device=device)
        anchors = anchors.view(num_layers, -1, 2) / strides.view(-1, 1, 1)
        targets_cls, targets_box, indices, anchors_encode = self.assign_targets_to_anchors(
            head_outputs, anchors, targets)

        return targets_cls, targets_box, indices, anchors_encode

    def assign_targets_to_anchors(
        self,
        head_outputs: List[Tensor],
        anchors: Tensor,
        targets: Tensor,
    ):
        """Assign ground truth boxes and targets to anchors.
        Args:
            gt_boxes (List[Tensor]): with shape num_targets x 4, ground truth boxes
            gt_labels (List[Tensor]): with shape num_targets, labels of targets
            anchors (Tensor): with shape num_priors x 4, XYXY_REL BoxMode
        Returns:
            boxes (List[Tensor]): with shape num_priors x 4 real values for anchors.
            labels (List[Tensor]): with shape num_priros, labels for anchors.
        """
        device = head_outputs[0].device
        num_layers = len(head_outputs)
        # Build targets for compute_loss(), input targets(image,class,x,y,w,h)
        num_anchors = len(self.anchor_grids)  # number of anchors
        num_targets = len(targets)  # number of targets

        targets_cls, targets_box, anchors_encode = [], [], []
        indices: List[Tuple[Tensor, Tensor, Tensor, Tensor]] = []
        gain = torch.ones(7, device=device)  # normalized to gridspace gain
        # same as .repeat_interleave(num_targets)
        ai = torch.arange(num_anchors, device=device).float().view(num_anchors, 1).repeat(1, num_targets)
        targets = torch.cat((targets.repeat(num_anchors, 1, 1), ai[:, :, None]), 2)  # append anchor indices

        g = 0.5  # bias
        off = torch.tensor([[0, 0],
                            [1, 0], [0, 1], [-1, 0], [0, -1],  # j,k,l,m
                            # [1, 1], [1, -1], [-1, 1], [-1, -1],  # jk,jm,lk,lm
                            ], device=device).float() * g  # offsets

        for i in range(num_layers):
            anchors_per_layer = anchors[i]
            gain[2:6] = torch.tensor(head_outputs[i].shape)[[3, 2, 3, 2]]  # xyxy gain

            # Match targets to anchors
            targets_with_gain = targets * gain
            if num_targets:
                # Matches
                ratios_wh = targets_with_gain[:, :, 4:6] / anchors_per_layer[:, None]  # wh ratio
                ratios_filtering = torch.max(ratios_wh, 1. / ratios_wh).max(2)[0]
                inds = torch.where(ratios_filtering < self.anchor_threshold)
                targets_with_gain = targets_with_gain[inds]  # filter

                # Offsets
                grid_xy = targets_with_gain[:, 2:4]  # grid xy
                grid_xy_inverse = gain[[2, 3]] - grid_xy  # inverse
                inds_jk = (grid_xy % 1. < g) & (grid_xy > 1.)
                inds_lm = (grid_xy_inverse % 1. < g) & (grid_xy_inverse > 1.)
                inds_ones = torch.ones_like(inds_jk[:, 0])[:, None]
                inds = torch.cat((inds_ones, inds_jk, inds_lm), dim=1).T
                targets_with_gain = targets_with_gain.repeat((5, 1, 1))[inds]
                offsets = (torch.zeros_like(grid_xy)[None] + off[:, None])[inds]
            else:
                targets_with_gain = targets[0]
                offsets = torch.tensor(0, device=device)

            # Define
            bc = targets_with_gain[:, :2].long().T  # image, class
            grid_xy = targets_with_gain[:, 2:4]  # grid xy
            grid_wh = targets_with_gain[:, 4:6]  # grid wh
            grid_ij = (grid_xy - offsets).long()

            # Append
            a = targets_with_gain[:, 6].long()  # anchor indices
            # image, anchor, grid indices
            indices.append((bc[0], a, grid_ij[:, 1].clamp_(0, gain[3] - 1), grid_ij[:, 0].clamp_(0, gain[2] - 1)))
            targets_box.append(torch.cat((grid_xy - grid_ij, grid_wh), 1))  # box
            anchors_encode.append(anchors_per_layer[a])  # anchors
            targets_cls.append(bc[1])  # class

        return targets_cls, targets_box, indices, anchors_encode

    @staticmethod
    def label_smooth_bce(eps: float = 0.1):
        '''
        Return positive, negative label smoothing BCE targets
        <https://github.com/ultralytics/yolov3/issues/238#issuecomment-598028441>
        '''
        return 1.0 - 0.5 * eps, 0.5 * eps

    def compute_loss(
        self,
        head_outputs: List[Tensor],
        targets_cls: List[Tensor],
        targets_box: List[Tensor],
        matched_idxs: List[Tuple[Tensor, Tensor, Tensor, Tensor]],
        anchors: List[Tensor],
    ) -> Dict[str, Tensor]:
        """ This performs the loss computation.
        Parameters:
            outputs: dict of tensors, see the output specification of the model for the format
            targets: list of dicts, such that len(targets) == batch_size.
                The expected keys in each dict depends on the losses applied, see each loss' doc
        """
        device = head_outputs[0].device
        num_classes = head_outputs[0].shape[-1] - 5
        loss_cls = torch.zeros(1, device=device)
        loss_box = torch.zeros(1, device=device)
        loss_obj = torch.zeros(1, device=device)

        # Class label smoothing https://arxiv.org/pdf/1902.04103.pdf eqn 3
        cls_positive, cls_negative = self.label_smooth_bce(eps=0.0)

        assert len(head_outputs) == len(self.layer_balance)
        # Losses
        num_targets = 0  # number of targets
        num_output = len(head_outputs)  # number of outputs

        cls_pw = torch.tensor([self.cls_pw], device=device)
        obj_pw = torch.tensor([self.obj_pw], device=device)

        for i, pred_logits_per_layer in enumerate(head_outputs):  # layer index, layer predictions
            b, a, gj, gi = matched_idxs[i]  # image, anchor, gridy, gridx
            obj_logits = torch.zeros_like(pred_logits_per_layer[..., 0], device=device)  # target obj

            num_target_per_layer = b.shape[0]  # number of targets
            if num_target_per_layer:
                num_targets += num_target_per_layer  # cumulative targets
                pred_logits_matched = pred_logits_per_layer[b, a, gj, gi]  # prediction subset corresponding to targets

                # Regression head
                bbox_xy = pred_logits_matched[:, :2].sigmoid() * 2. - 0.5
                bbox_wh = (pred_logits_matched[:, 2:4].sigmoid() * 2) ** 2 * anchors[i]
                bbox_regression = torch.cat((bbox_xy, bbox_wh), 1).to(device)  # predicted box
                ciou = det_utils.bbox_ciou(bbox_regression.T, targets_box[i])
                loss_box += (1.0 - ciou).mean()  # iou loss

                # Objectness head
                # iou ratio
                ciou_vals = torch.tensor(ciou.detach().clamp(0), dtype=obj_logits.dtype)
                obj_logits[b, a, gj, gi] = (1.0 - self.iou_ratio) + (self.iou_ratio * ciou_vals)

                # Classification head
                if num_classes > 1:  # cls loss (only if multiple classes)
                    cls_logits = torch.full_like(pred_logits_matched[:, 5:], cls_negative, device=device)  # targets
                    cls_logits[torch.arange(num_target_per_layer), targets_cls[i]] = cls_positive

                    loss_cls += det_utils.cls_loss(pred_logits_matched[:, 5:], cls_logits, pos_weight=cls_pw)  # BCE

            loss_obj += det_utils.obj_loss(
                pred_logits_per_layer[..., 4],
                obj_logits,
                pos_weight=obj_pw,
            ) * self.layer_balance[i]  # obj loss

        out_scaling = 3 / num_output  # output count scaling
        loss_box *= self.box * out_scaling
        loss_obj *= self.obj * out_scaling * (1.4 if num_output == 4 else 1.)
        loss_cls *= self.cls * out_scaling

        return {
            'cls_logits': loss_cls,
            'bbox_regression': loss_box,
            'objectness': loss_obj,
        }


class PostProcess(nn.Module):
    __annotations__ = {
        'box_coder': det_utils.BoxCoder,
    }
    """Performs Non-Maximum Suppression (NMS) on inference results"""
    def __init__(
        self,
        score_thresh: float,
        nms_thresh: float,
        detections_per_img: int,
    ) -> None:
        """
        Arguments:
            score_thresh (float)
            nms_thresh (float)
            detections_per_img (int)
        """
        super().__init__()
        self.box_coder = det_utils.BoxCoder()
        self.score_thresh = score_thresh
        self.nms_thresh = nms_thresh
        self.detections_per_img = detections_per_img  # maximum number of detections per image

    def forward(
        self,
        head_outputs: List[Tensor],
        anchors_tuple: Tuple[Tensor, Tensor, Tensor],
    ) -> List[Dict[str, Tensor]]:
        """ Perform the computation. At test time, postprocess_detections is the final layer of YOLO.
        Decode location preds, apply non-maximum suppression to location predictions based on conf
        scores and threshold to a detections_per_img number of output predictions for both confidence
        score and locations.

        Parameters:
            head_outputs : [batch_size, num_anchors, num_classes + 5] predicted locations and class/object confidence.
            image_shapes: tensor of dimension [batch_size x 2] containing the size of each images of the batch
                          For evaluation, this must be the original image size (before any data augmentation)
                          For visualization, this should be the image size after data augment, but before padding
        """
        batch_size, _, _, _, K = head_outputs[0].shape

        all_pred_logits: List[Tensor] = []
        for pred_logits in head_outputs:
            pred_logits = pred_logits.reshape(batch_size, -1, K)  # Size=(NN, HWA, K)
            all_pred_logits.append(pred_logits)

        all_pred_logits = torch.cat(all_pred_logits, dim=1)

        detections: List[Dict[str, Tensor]] = []

        for idx in range(batch_size):  # image idx, image inference
            pred_logits = torch.sigmoid(all_pred_logits[idx])

            # Compute conf
            # box_conf x class_conf, w/ shape: num_anchors x num_classes
            scores = pred_logits[:, 5:] * pred_logits[:, 4:5]

            boxes = self.box_coder.decode_single(pred_logits[:, :4], anchors_tuple)

            # remove low scoring boxes
            inds, labels = torch.where(scores > self.score_thresh)
            boxes, scores = boxes[inds], scores[inds, labels]

            # non-maximum suppression, independently done per level
            keep = batched_nms(boxes, scores, labels, self.nms_thresh)
            # keep only topk scoring head_outputs
            keep = keep[:self.detections_per_img]
            boxes, scores, labels = boxes[keep], scores[keep], labels[keep]

            detections.append({'scores': scores, 'labels': labels, 'boxes': boxes})

        return detections



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
# Modified by Zhiqiang Wang (zhiqwang@outlook.com)
import warnings

import torch
from torch import nn, Tensor

from torchvision.models.utils import load_state_dict_from_url

from .backbone_utils import darknet_pan_backbone
from .anchor_utils import AnchorGenerator
from .box_head import YoloHead, SetCriterion, PostProcess

from typing import Tuple, Any, List, Dict, Optional

__all__ = ['YOLO', 'yolov5_darknet_pan_s_r31', 'yolov5_darknet_pan_m_r31', 'yolov5_darknet_pan_l_r31',
           'yolov5_darknet_pan_s_r40', 'yolov5_darknet_pan_m_r40', 'yolov5_darknet_pan_l_r40']


class YOLO(nn.Module):
    def __init__(
        self,
        backbone: nn.Module,
        num_classes: int,
        anchor_grids: List[List[int]],
        # Anchor parameters
        anchor_generator: Optional[nn.Module] = None,
        head: Optional[nn.Module] = None,
        # Training parameter
        loss_calculator: Optional[nn.Module] = None,
        # Post Process parameter
        post_process: Optional[nn.Module] = None,
        score_thresh: float = 0.05,
        nms_thresh: float = 0.5,
        detections_per_img: int = 300,
    ):
        super().__init__()
        if not hasattr(backbone, "out_channels"):
            raise ValueError(
                "backbone should contain an attribute out_channels "
                "specifying the number of output channels (assumed to be the "
                "same for all the levels)")
        self.backbone = backbone

        if anchor_generator is None:
            strides: List[int] = [8, 16, 32]
            anchor_generator = AnchorGenerator(strides, anchor_grids)
        self.anchor_generator = anchor_generator

        if loss_calculator is None:
            strides: List[int] = [8, 16, 32]
            loss_calculator = SetCriterion(strides, anchor_grids)
        self.compute_loss = loss_calculator

        if head is None:
            head = YoloHead(
                backbone.out_channels,
                anchor_generator.num_anchors,
                num_classes,
            )
        self.head = head

        if post_process is None:
            post_process = PostProcess(score_thresh, nms_thresh, detections_per_img)
        self.post_process = post_process

        # used only on torchscript mode
        self._has_warned = False

    @torch.jit.unused
    def eager_outputs(
        self,
        losses: Dict[str, Tensor],
        detections: List[Dict[str, Tensor]],
    ) -> Tuple[Dict[str, Tensor], List[Dict[str, Tensor]]]:
        if self.training:
            return losses

        return detections

    def forward(
        self,
        samples: Tensor,
        targets: Optional[Tensor] = None,
    ) -> Tuple[Dict[str, Tensor], List[Dict[str, Tensor]]]:
        """
        Arguments:
            samples (NestedTensor): Expects a NestedTensor, which consists of:
               - samples.tensor: batched images, of shape [batch_size x 3 x H x W]
            targets (list[Dict[Tensor]]): ground-truth boxes present in the image (optional)

        Returns:
            result (list[BoxList] or dict[Tensor]): the output from the model.
                During training, it returns a dict[Tensor] which contains the losses.
                During testing, it returns list[BoxList] contains additional fields
                like `scores`, `labels` and `mask` (for Mask R-CNN models).
        """
        # get the features from the backbone
        features = self.backbone(samples)

        # compute the yolo heads outputs using the features
        head_outputs = self.head(features)

        # create the set of anchors
        anchors_tuple = self.anchor_generator(features)
        losses = {}
        detections: List[Dict[str, Tensor]] = []

        if self.training:
            assert targets is not None
            # compute the losses
            losses = self.compute_loss(head_outputs, targets)
        else:
            # compute the detections
            detections = self.post_process(head_outputs, anchors_tuple)

        if torch.jit.is_scripting():
            if not self._has_warned:
                warnings.warn("YOLO always returns a (Losses, Detections) tuple in scripting")
                self._has_warned = True
            return losses, detections
        else:
            return self.eager_outputs(losses, detections)


model_urls_root = 'https://github.com/zhiqwang/yolov5-rt-stack/releases/download/v0.3.0'

model_urls = {
    'yolov5_darknet_pan_s_r31_coco': f'{model_urls_root}/yolov5_darknet_pan_s_r31_coco-eb728698.pt',
    'yolov5_darknet_pan_m_r31_coco': f'{model_urls_root}/yolov5_darknet_pan_m_r31_coco-670dc553.pt',
    'yolov5_darknet_pan_l_r31_coco': f'{model_urls_root}/yolov5_darknet_pan_l_r31_coco-4dcc8209.pt',
    'yolov5_darknet_pan_s_r40_coco': None,
    'yolov5_darknet_pan_m_r40_coco': None,
    'yolov5_darknet_pan_l_r40_coco': None,
}


def _yolov5_darknet_pan(
    backbone_name: str,
    depth_multiple: float,
    width_multiple: float,
    weights_name: str,
    pretrained: bool = False,
    progress: bool = True,
    num_classes: int = 80,
    **kwargs: Any,
) -> YOLO:
    """
    Constructs a YOLO model.

    The input to the model is expected to be a batched tensors, of shape ``[N, C, H, W]``, one for each
    image, and should be in ``0-1`` range. Different images can have different sizes.

    The behavior of the model changes depending if it is in training or evaluation mode.

    During training, the model expects both the input tensors, as well as a targets (list of dictionary),
    containing:
        - boxes (``FloatTensor[N, 4]``): the ground-truth boxes in ``[x1, y1, x2, y2]`` format, with values
          between ``0`` and ``H`` and ``0`` and ``W``
        - labels (``Int64Tensor[N]``): the class label for each ground-truth box

    The model returns a ``Dict[Tensor]`` during training, containing the classification and regression
    losses.

    During inference, the model requires only the input tensors, and returns the post-processed
    predictions as a ``List[Dict[Tensor]]``, one for each input image. The fields of the ``Dict`` are as
    follows:
        - boxes (``FloatTensor[N, 4]``): the predicted boxes in ``[x1, y1, x2, y2]`` format, with values between
          ``0`` and ``H`` and ``0`` and ``W``
        - labels (``Int64Tensor[N]``): the predicted labels for each image
        - scores (``Tensor[N]``): the scores or each prediction

    Example::

        >>> model = yolov5(pretrained=True)
        >>> model.eval()
        >>> x = torch.rand(4, 3, 416, 320)
        >>> predictions = model(x)

    Arguments:
        pretrained (bool): If True, returns a model pre-trained on COCO train2017
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    backbone = darknet_pan_backbone(backbone_name, depth_multiple, width_multiple)

    anchor_grids = [[10, 13, 16, 30, 33, 23],
                    [30, 61, 62, 45, 59, 119],
                    [116, 90, 156, 198, 373, 326]]

    model = YOLO(backbone, num_classes, anchor_grids, **kwargs)
    if pretrained:
        if model_urls.get(weights_name, None) is None:
            raise ValueError(f"No checkpoint is available for model {weights_name}")
        state_dict = load_state_dict_from_url(model_urls[weights_name], progress=progress)
        model.load_state_dict(state_dict)

    return model


def yolov5_darknet_pan_s_r31(pretrained: bool = False, progress: bool = True, num_classes: int = 80,
                             **kwargs: Any) -> YOLO:
    r"""yolov5 small release 3.1 model from
    `"ultralytics/yolov5" <https://zenodo.org/badge/latestdoi/264818686>`_.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    backbone_name = 'darknet_s_r3_1'
    weights_name = 'yolov5_darknet_pan_s_r31_coco'
    depth_multiple = 0.33
    width_multiple = 0.5
    return _yolov5_darknet_pan(backbone_name, depth_multiple, width_multiple, weights_name,
                               pretrained=pretrained, progress=progress, num_classes=num_classes, **kwargs)


def yolov5_darknet_pan_m_r31(pretrained: bool = False, progress: bool = True, num_classes: int = 80,
                             **kwargs: Any) -> YOLO:
    r"""yolov5 medium release 3.1 model from
    `"ultralytics/yolov5" <https://zenodo.org/badge/latestdoi/264818686>`_.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    backbone_name = 'darknet_m_r3_1'
    weights_name = 'yolov5_darknet_pan_m_r31_coco'
    depth_multiple = 0.67
    width_multiple = 0.75
    return _yolov5_darknet_pan(backbone_name, depth_multiple, width_multiple, weights_name,
                               pretrained=pretrained, progress=progress, num_classes=num_classes, **kwargs)


def yolov5_darknet_pan_l_r31(pretrained: bool = False, progress: bool = True, num_classes: int = 80,
                             **kwargs: Any) -> YOLO:
    r"""yolov5 large release 3.1 model from
    `"ultralytics/yolov5" <https://zenodo.org/badge/latestdoi/264818686>`_.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    backbone_name = 'darknet_l_r3_1'
    weights_name = 'yolov5_darknet_pan_l_r31_coco'
    depth_multiple = 1.0
    width_multiple = 1.0
    return _yolov5_darknet_pan(backbone_name, depth_multiple, width_multiple, weights_name,
                               pretrained=pretrained, progress=progress, num_classes=num_classes, **kwargs)


def yolov5_darknet_pan_s_r40(pretrained: bool = False, progress: bool = True, num_classes: int = 80,
                             **kwargs: Any) -> YOLO:
    r"""yolov5 small release 4.0 model from
    `"ultralytics/yolov5" <https://zenodo.org/badge/latestdoi/264818686>`_.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    backbone_name = 'darknet_s_r4_0'
    weights_name = 'yolov5_darknet_pan_s_r40_coco'
    depth_multiple = 0.33
    width_multiple = 0.5
    return _yolov5_darknet_pan(backbone_name, depth_multiple, width_multiple, weights_name,
                               pretrained=pretrained, progress=progress, num_classes=num_classes, **kwargs)


def yolov5_darknet_pan_m_r40(pretrained: bool = False, progress: bool = True, num_classes: int = 80,
                             **kwargs: Any) -> YOLO:
    r"""yolov5 medium release 4.0 model from
    `"ultralytics/yolov5" <https://zenodo.org/badge/latestdoi/264818686>`_.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    backbone_name = 'darknet_m_r4_0'
    weights_name = 'yolov5_darknet_pan_m_r40_coco'
    depth_multiple = 0.67
    width_multiple = 0.75
    return _yolov5_darknet_pan(backbone_name, depth_multiple, width_multiple, weights_name,
                               pretrained=pretrained, progress=progress, num_classes=num_classes, **kwargs)


def yolov5_darknet_pan_l_r40(pretrained: bool = False, progress: bool = True, num_classes: int = 80,
                             **kwargs: Any) -> YOLO:
    r"""yolov5 large release 4.0 model from
    `"ultralytics/yolov5" <https://zenodo.org/badge/latestdoi/264818686>`_.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    backbone_name = 'darknet_l_r4_0'
    weights_name = 'yolov5_darknet_pan_l_r40_coco'
    depth_multiple = 1.0
    width_multiple = 1.0
    return _yolov5_darknet_pan(backbone_name, depth_multiple, width_multiple, weights_name,
                               pretrained=pretrained, progress=progress, num_classes=num_classes, **kwargs)



import io
import torch

# onnxruntime requires python 3.5 or above
try:
    import onnxruntime
except ImportError:
    onnxruntime = None

import unittest
from torchvision.ops._register_onnx_ops import _onnx_opset_version

from yolort.models import yolov5_onnx


@unittest.skipIf(onnxruntime is None, 'ONNX Runtime unavailable')
class ONNXExporterTester(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        torch.manual_seed(123)

    def run_model(self, model, inputs_list, tolerate_small_mismatch=False, do_constant_folding=True, dynamic_axes=None,
                  output_names=None, input_names=None):
        model.eval()

        onnx_io = io.BytesIO()
        # export to onnx with the first input
        torch.onnx.export(model, inputs_list[0], onnx_io,
                          do_constant_folding=do_constant_folding, opset_version=_onnx_opset_version,
                          dynamic_axes=dynamic_axes, input_names=input_names, output_names=output_names)
        # validate the exported model with onnx runtime
        for test_inputs in inputs_list:
            with torch.no_grad():
                if isinstance(test_inputs, torch.Tensor) or isinstance(test_inputs, list):
                    test_inputs = (test_inputs,)
                test_ouputs = model(*test_inputs)
                if isinstance(test_ouputs, torch.Tensor):
                    test_ouputs = (test_ouputs,)
            self.ort_validate(onnx_io, test_inputs, test_ouputs, tolerate_small_mismatch)

    def ort_validate(self, onnx_io, inputs, outputs, tolerate_small_mismatch=False):

        inputs, _ = torch.jit._flatten(inputs)
        outputs, _ = torch.jit._flatten(outputs)

        def to_numpy(tensor):
            if tensor.requires_grad:
                return tensor.detach().cpu().numpy()
            else:
                return tensor.cpu().numpy()

        inputs = list(map(to_numpy, inputs))
        outputs = list(map(to_numpy, outputs))

        ort_session = onnxruntime.InferenceSession(onnx_io.getvalue())
        # compute onnxruntime output prediction
        ort_inputs = dict((ort_session.get_inputs()[i].name, inpt) for i, inpt in enumerate(inputs))
        ort_outs = ort_session.run(None, ort_inputs)

        for i in range(0, len(outputs)):
            try:
                torch.testing.assert_allclose(outputs[i], ort_outs[i], rtol=1e-03, atol=1e-05)
            except AssertionError as error:
                if tolerate_small_mismatch:
                    self.assertIn("(0.00%)", str(error), str(error))
                else:
                    raise

    def get_image_from_url(self, url, size=None):
        import requests
        from PIL import Image
        from io import BytesIO
        from torchvision import transforms

        data = requests.get(url)
        image = Image.open(BytesIO(data.content)).convert("RGB")

        if size is None:
            size = (300, 200)
        image = image.resize(size, Image.BILINEAR)

        to_tensor = transforms.ToTensor()
        return to_tensor(image)

    def get_test_images(self):
        image_url = "http://farm3.staticflickr.com/2469/3915380994_2e611b1779_z.jpg"
        image = self.get_image_from_url(url=image_url, size=(100, 320))

        image_url2 = "https://pytorch.org/tutorials/_static/img/tv_tutorial/tv_image05.png"
        image2 = self.get_image_from_url(url=image_url2, size=(250, 380))

        images = [image]
        test_images = [image2]
        return images, test_images

    def test_yolov5s(self):
        images, test_images = self.get_test_images()
        dummy_image = [torch.ones(3, 100, 100) * 0.3]
        model = yolov5_onnx(pretrained=True)
        model.eval()
        model(images)
        # Test exported model on images of different size, or dummy input
        self.run_model(model, [(images,), (test_images,), (dummy_image,)], input_names=["images_tensors"],
                       output_names=["outputs"],
                       dynamic_axes={"images_tensors": [0, 1, 2], "outputs": [0, 1, 2]},
                       tolerate_small_mismatch=True)
        # Test exported model for an image with no detections on other images
        self.run_model(model, [(dummy_image,), (images,)], input_names=["images_tensors"],
                       output_names=["outputs"],
                       dynamic_axes={"images_tensors": [0, 1, 2], "outputs": [0, 1, 2]},
                       tolerate_small_mismatch=True)


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2021, Zhiqiang Wang. All Rights Reserved.
from typing import List, Optional
from torch import nn
from torchvision.models._utils import IntermediateLayerGetter

from . import darknet
from .path_aggregation_network import PathAggregationNetwork


class BackboneWithPAN(nn.Module):
    """
    Adds a PAN on top of a model.
    Internally, it uses torchvision.models._utils.IntermediateLayerGetter to
    extract a submodel that returns the feature maps specified in return_layers.
    The same limitations of IntermediatLayerGetter apply here.
    Args:
        backbone (nn.Module)
        return_layers (Dict[name, new_name]): a dict containing the names
            of the modules for which the activations will be returned as
            the key of the dict, and the value of the dict is the name
            of the returned activation (which the user can specify).
        in_channels_list (List[int]): number of channels for each feature map
            that is returned, in the order they are present in the OrderedDict
    Attributes:
        out_channels (int): the number of channels in the PAN
    """
    def __init__(self, backbone, return_layers, in_channels_list, depth_multiple):
        super().__init__()

        self.body = IntermediateLayerGetter(backbone, return_layers=return_layers)
        self.pan = PathAggregationNetwork(in_channels_list, depth_multiple)
        self.out_channels = in_channels_list

    def forward(self, x):
        x = self.body(x)
        x = self.pan(x)
        return x


def darknet_pan_backbone(
    backbone_name: str,
    depth_multiple: float,
    width_multiple: float,
    pretrained: Optional[bool] = False,
    returned_layers: Optional[List[int]] = None,
):
    """
    Constructs a specified ResNet backbone with PAN on top. Freezes the specified number of layers in the backbone.

    Examples::

        >>> from models.backbone_utils import darknet_pan_backbone
        >>> backbone = darknet_pan_backbone('darknet3_1', pretrained=True, trainable_layers=3)
        >>> # get some dummy image
        >>> x = torch.rand(1, 3, 64, 64)
        >>> # compute the output
        >>> output = backbone(x)
        >>> print([(k, v.shape) for k, v in output.items()])
        >>> # returns
        >>>   [('0', torch.Size([1, 128, 8, 8])),
        >>>    ('1', torch.Size([1, 256, 4, 4])),
        >>>    ('2', torch.Size([1, 512, 2, 2]))]

    Args:
        backbone_name (string): resnet architecture. Possible values are 'ResNet', 'resnet18', 'resnet34', 'resnet50',
             'resnet101', 'resnet152', 'resnext50_32x4d', 'resnext101_32x8d', 'wide_resnet50_2', 'wide_resnet101_2'
        norm_layer (torchvision.ops): it is recommended to use the default value. For details visit:
            (https://github.com/facebookresearch/maskrcnn-benchmark/issues/267)
        pretrained (bool): If True, returns a model with backbone pre-trained on Imagenet
        trainable_layers (int): number of trainable (not frozen) resnet layers starting from final block.
            Valid values are between 0 and 5, with 5 meaning all backbone layers are trainable.
    """
    backbone = darknet.__dict__[backbone_name](pretrained=pretrained).features

    if returned_layers is None:
        returned_layers = [4, 6, 8]

    return_layers = {str(k): str(i) for i, k in enumerate(returned_layers)}

    in_channels_list = [int(gw * width_multiple) for gw in [256, 512, 1024]]

    return BackboneWithPAN(backbone, return_layers, in_channels_list, depth_multiple)



# Copyright (c) 2021, Zhiqiang Wang. All Rights Reserved.
from pathlib import Path

import torch.utils.data
from torch.utils.data import DataLoader
from torch.utils.data.dataset import Dataset

from pytorch_lightning import LightningDataModule

from .transforms import collate_fn, default_train_transforms, default_val_transforms
from .voc import VOCDetection
from .coco import CocoDetection

from typing import Callable, List, Any, Optional


class DetectionDataModule(LightningDataModule):
    """
    Wrapper of Datasets in LightningDataModule
    """
    def __init__(
        self,
        train_dataset: Optional[Dataset] = None,
        val_dataset: Optional[Dataset] = None,
        test_dataset: Optional[Dataset] = None,
        batch_size: int = 1,
        num_workers: int = 0,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)

        self._train_dataset = train_dataset
        self._val_dataset = val_dataset
        self._test_dataset = test_dataset

        self.batch_size = batch_size
        self.num_workers = num_workers

    def train_dataloader(self, batch_size: int = 16) -> None:
        """
        VOCDetection and CocoDetection
        Args:
            batch_size: size of batch
            transforms: custom transforms
        """
        # Creating data loaders
        sampler = torch.utils.data.RandomSampler(self._train_dataset)
        batch_sampler = torch.utils.data.BatchSampler(sampler, batch_size, drop_last=True)

        loader = DataLoader(
            self._train_dataset,
            batch_sampler=batch_sampler,
            collate_fn=collate_fn,
            num_workers=self.num_workers,
        )

        return loader

    def val_dataloader(self, batch_size: int = 16) -> None:
        """
        VOCDetection and CocoDetection
        Args:
            batch_size: size of batch
            transforms: custom transforms
        """
        # Creating data loaders
        sampler = torch.utils.data.SequentialSampler(self._val_dataset)

        loader = DataLoader(
            self._val_dataset,
            batch_size,
            sampler=sampler,
            drop_last=False,
            collate_fn=collate_fn,
            num_workers=self.num_workers,
        )

        return loader


class VOCDetectionDataModule(DetectionDataModule):
    def __init__(
        self,
        data_path: str,
        years: List[str] = ["2007", "2012"],
        train_transform: Optional[Callable] = default_train_transforms,
        val_transform: Optional[Callable] = default_val_transforms,
        batch_size: int = 1,
        num_workers: int = 0,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        train_dataset, num_classes = self.build_datasets(
            data_path, image_set='train', years=years, transforms=train_transform)
        val_dataset, _ = self.build_datasets(
            data_path, image_set='val', years=years, transforms=val_transform)

        super().__init__(train_dataset=train_dataset, val_dataset=val_dataset,
                         batch_size=batch_size, num_workers=num_workers, *args, **kwargs)

        self.num_classes = num_classes

    @staticmethod
    def build_datasets(data_path, image_set, years, transforms):
        datasets = []
        for year in years:
            dataset = VOCDetection(
                data_path,
                year=year,
                image_set=image_set,
                transforms=transforms(),
            )
            datasets.append(dataset)

        num_classes = len(datasets[0].prepare.CLASSES)

        if len(datasets) == 1:
            return datasets[0], num_classes
        else:
            return torch.utils.data.ConcatDataset(datasets), num_classes


class CocoDetectionDataModule(DetectionDataModule):
    def __init__(
        self,
        data_path: str,
        year: str = "2017",
        train_transform: Optional[Callable] = default_train_transforms,
        val_transform: Optional[Callable] = default_val_transforms,
        batch_size: int = 1,
        num_workers: int = 0,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        train_dataset = self.build_datasets(
            data_path, image_set='train', year=year, transforms=train_transform)
        val_dataset = self.build_datasets(
            data_path, image_set='val', year=year, transforms=val_transform)

        super().__init__(train_dataset=train_dataset, val_dataset=val_dataset,
                         batch_size=batch_size, num_workers=num_workers, *args, **kwargs)

        self.num_classes = 80

    @staticmethod
    def build_datasets(data_path, image_set, year, transforms):
        ann_file = Path(data_path).joinpath('annotations').joinpath(f"instances_{image_set}{year}.json")
        return CocoDetection(data_path, ann_file, transforms())



# Copyright (c) 2021, Zhiqiang Wang. All Rights Reserved.

from yolort import models
from yolort import datasets
from yolort import utils

try:
    from .version import __version__  # noqa: F401
except ImportError:
    pass



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
"""
COCO dataset which returns image_id for evaluation.
Mostly copy-paste from https://github.com/pytorch/vision/blob/13b35ff/references/detection/coco_utils.py
"""
from pathlib import Path

import torch
import torch.utils.data
import torchvision
from pycocotools import mask as coco_mask


class ConvertCocoPolysToMask(object):
    def __init__(self, json_category_id_maps, return_masks=False):
        self.json_category_id_to_contiguous_id = json_category_id_maps
        self.return_masks = return_masks

    def __call__(self, image, target):
        w, h = image.size

        image_id = target["image_id"]
        image_id = torch.tensor([image_id])

        anno = target["annotations"]

        anno = [obj for obj in anno if 'iscrowd' not in obj or obj['iscrowd'] == 0]

        boxes = [obj["bbox"] for obj in anno]
        # guard against no boxes via resizing
        boxes = torch.as_tensor(boxes, dtype=torch.float32).reshape(-1, 4)
        # BoxMode: convert from XYWH_ABS to XYXY_ABS
        boxes[:, 2:] += boxes[:, :2]
        boxes[:, 0::2].clamp_(min=0, max=w)
        boxes[:, 1::2].clamp_(min=0, max=h)

        classes = [obj["category_id"] for obj in anno]
        classes = [self.json_category_id_to_contiguous_id[c] for c in classes]
        classes = torch.tensor(classes, dtype=torch.int64)

        if self.return_masks:
            segmentations = [obj["segmentation"] for obj in anno]
            masks = convert_coco_poly_to_mask(segmentations, h, w)

        keypoints = None
        if anno and "keypoints" in anno[0]:
            keypoints = [obj["keypoints"] for obj in anno]
            keypoints = torch.as_tensor(keypoints, dtype=torch.float32)
            num_keypoints = keypoints.shape[0]
            if num_keypoints:
                keypoints = keypoints.view(num_keypoints, -1, 3)

        keep = (boxes[:, 3] > boxes[:, 1]) & (boxes[:, 2] > boxes[:, 0])
        boxes = boxes[keep]
        classes = classes[keep]
        if self.return_masks:
            masks = masks[keep]
        if keypoints is not None:
            keypoints = keypoints[keep]

        target = {}
        target["boxes"] = boxes
        target["labels"] = classes
        if self.return_masks:
            target["masks"] = masks
        target["image_id"] = image_id
        if keypoints is not None:
            target["keypoints"] = keypoints

        # for conversion to coco api
        area = torch.tensor([obj["area"] for obj in anno])
        iscrowd = torch.tensor([obj["iscrowd"] if "iscrowd" in obj else 0 for obj in anno])
        target["area"] = area[keep]
        target["iscrowd"] = iscrowd[keep]

        target["orig_size"] = torch.as_tensor([int(h), int(w)])
        target["size"] = torch.as_tensor([int(h), int(w)])

        return image, target


class CocoDetection(torchvision.datasets.CocoDetection):
    def __init__(self, img_folder, ann_file, transforms, return_masks=False):
        super().__init__(img_folder, ann_file)
        self._transforms = transforms

        json_category_id_to_contiguous_id = {
            v: i for i, v in enumerate(self.coco.getCatIds())
        }
        self.prepare = ConvertCocoPolysToMask(json_category_id_to_contiguous_id, return_masks)

    def __getitem__(self, idx):
        img, target = super().__getitem__(idx)
        image_id = self.ids[idx]
        target = {'image_id': image_id, 'annotations': target}
        img, target = self.prepare(img, target)
        if self._transforms is not None:
            img, target = self._transforms(img, target)
        return img, target


def convert_coco_poly_to_mask(segmentations, height, width):
    masks = []
    for polygons in segmentations:
        rles = coco_mask.frPyObjects(polygons, height, width)
        mask = coco_mask.decode(rles)
        if len(mask.shape) < 3:
            mask = mask[..., None]
        mask = torch.as_tensor(mask, dtype=torch.uint8)
        mask = mask.any(dim=2)
        masks.append(mask)
    if masks:
        masks = torch.stack(masks, dim=0)
    else:
        masks = torch.zeros((0, height, width), dtype=torch.uint8)
    return masks


def _coco_remove_images_without_annotations(dataset, cat_list=None):
    def _has_only_empty_bbox(anno):
        return all(any(o <= 1 for o in obj["bbox"][2:]) for obj in anno)

    def _count_visible_keypoints(anno):
        return sum(sum(1 for v in ann["keypoints"][2::3] if v > 0) for ann in anno)

    min_keypoints_per_image = 10

    def _has_valid_annotation(anno):
        # if it's empty, there is no annotation
        if len(anno) == 0:
            return False
        # if all boxes have close to zero area, there is no annotation
        if _has_only_empty_bbox(anno):
            return False
        # keypoints task have a slight different critera for considering
        # if an annotation is valid
        if "keypoints" not in anno[0]:
            return True
        # for keypoint detection tasks, only consider valid images those
        # containing at least min_keypoints_per_image
        if _count_visible_keypoints(anno) >= min_keypoints_per_image:
            return True
        return False

    assert isinstance(dataset, torchvision.datasets.CocoDetection)
    ids = []
    for ds_idx, img_id in enumerate(dataset.ids):
        ann_ids = dataset.coco.getAnnIds(imgIds=img_id, iscrowd=None)
        anno = dataset.coco.loadAnns(ann_ids)
        if cat_list:
            anno = [obj for obj in anno if obj["category_id"] in cat_list]
        if _has_valid_annotation(anno):
            ids.append(ds_idx)

    dataset = torch.utils.data.Subset(dataset, ids)
    return dataset



# This file contains modules common to various models
import math
from typing import List

import torch
from torch import nn, Tensor


def autopad(k, p=None):  # kernel, padding
    # Pad to 'same'
    if p is None:
        p = k // 2 if isinstance(k, int) else [x // 2 for x in k]  # auto-pad
    return p


def DWConv(c1, c2, k=1, s=1, act=True):
    # Depthwise convolution
    return Conv(c1, c2, k, s, g=math.gcd(c1, c2), act=act)


class Conv(nn.Module):
    # Standard convolution
    def __init__(self, c1, c2, k=1, s=1, p=None, g=1, act=True):  # ch_in, ch_out, kernel, stride, padding, groups
        super().__init__()
        self.conv = nn.Conv2d(c1, c2, k, s, autopad(k, p), groups=g, bias=False)
        self.bn = nn.BatchNorm2d(c2)
        self.act = nn.Hardswish() if act else nn.Identity()

    def forward(self, x: Tensor) -> Tensor:
        return self.act(self.bn(self.conv(x)))

    def fuseforward(self, x):
        return self.act(self.conv(x))


class Bottleneck(nn.Module):
    # Standard bottleneck
    def __init__(self, c1, c2, shortcut=True, g=1, e=0.5):  # ch_in, ch_out, shortcut, groups, expansion
        super().__init__()
        c_ = int(c2 * e)  # hidden channels
        self.cv1 = Conv(c1, c_, 1, 1)
        self.cv2 = Conv(c_, c2, 3, 1, g=g)
        self.add = shortcut and c1 == c2

    def forward(self, x):
        return x + self.cv2(self.cv1(x)) if self.add else self.cv2(self.cv1(x))


class BottleneckCSP(nn.Module):
    # CSP Bottleneck https://github.com/WongKinYiu/CrossStagePartialNetworks
    def __init__(self, c1, c2, n=1, shortcut=True, g=1, e=0.5):  # ch_in, ch_out, number, shortcut, groups, expansion
        super().__init__()
        c_ = int(c2 * e)  # hidden channels
        self.cv1 = Conv(c1, c_, 1, 1)
        self.cv2 = nn.Conv2d(c1, c_, 1, 1, bias=False)
        self.cv3 = nn.Conv2d(c_, c_, 1, 1, bias=False)
        self.cv4 = Conv(2 * c_, c2, 1, 1)
        self.bn = nn.BatchNorm2d(2 * c_)  # applied to cat(cv2, cv3)
        self.act = nn.LeakyReLU(0.1, inplace=True)
        self.m = nn.Sequential(*[Bottleneck(c_, c_, shortcut, g, e=1.0) for _ in range(n)])

    def forward(self, x):
        y1 = self.cv3(self.m(self.cv1(x)))
        y2 = self.cv2(x)
        return self.cv4(self.act(self.bn(torch.cat((y1, y2), dim=1))))


class SPP(nn.Module):
    # Spatial pyramid pooling layer used in YOLOv3-SPP
    def __init__(self, c1, c2, k=(5, 9, 13)):
        super().__init__()
        c_ = c1 // 2  # hidden channels
        self.cv1 = Conv(c1, c_, 1, 1)
        self.cv2 = Conv(c_ * (len(k) + 1), c2, 1, 1)
        self.m = nn.ModuleList([nn.MaxPool2d(kernel_size=x, stride=1, padding=x // 2) for x in k])

    def forward(self, x):
        x = self.cv1(x)
        return self.cv2(torch.cat([x] + [m(x) for m in self.m], 1))


class Focus(nn.Module):
    # Focus wh information into c-space
    def __init__(self, c1, c2, k=1, s=1, p=None, g=1, act=True):  # ch_in, ch_out, kernel, stride, padding, groups
        super().__init__()
        self.conv = Conv(c1 * 4, c2, k, s, p, g, act)

    def forward(self, x: Tensor) -> Tensor:
        y = focus_transform(x)
        out = self.conv(y)

        return out


def focus_transform(x: Tensor) -> Tensor:
    '''x(b,c,w,h) -> y(b,4c,w/2,h/2)'''
    y = torch.cat([x[..., ::2, ::2], x[..., 1::2, ::2], x[..., ::2, 1::2], x[..., 1::2, 1::2]], 1)
    return y


class Concat(nn.Module):
    # Concatenate a list of tensors along dimension
    def __init__(self, dimension: int = 1):
        super().__init__()
        self.d = dimension

    # torchscript does not yet support *args, so we overload method
    # allowing it to take either a List[Tensor] or single Tensor
    def forward(self, x: List[Tensor]) -> Tensor:
        if isinstance(x, Tensor):
            prev_features = [x]
        else:
            prev_features = x
        return torch.cat(prev_features, self.d)


class Flatten(nn.Module):
    # Use after nn.AdaptiveAvgPool2d(1) to remove last 2 dimensions
    @staticmethod
    def forward(x):
        return x.view(x.size(0), -1)


class Classify(nn.Module):
    # Classification head, i.e. x(b,c1,20,20) to x(b,c2)
    def __init__(self, c1, c2, k=1, s=1, p=None, g=1):  # ch_in, ch_out, kernel, stride, padding, groups
        super().__init__()
        self.aap = nn.AdaptiveAvgPool2d(1)  # to x(b,c1,1,1)
        self.conv = nn.Conv2d(c1, c2, k, s, autopad(k, p), groups=g, bias=False)  # to x(b,c2,1,1)
        self.flat = Flatten()

    def forward(self, x):
        z = torch.cat([self.aap(y) for y in (x if isinstance(x, list) else [x])], 1)  # cat if list
        return self.flat(self.conv(z))  # flatten to x(b,c2)



# Copyright (c) 2020, Zhiqiang Wang. All Rights Reserved.
import torch
from torch import nn, Tensor
from typing import Tuple, List


class AnchorGenerator(nn.Module):

    def __init__(
        self,
        strides: List[int],
        anchor_grids: List[List[int]],
    ):
        super().__init__()
        assert len(strides) == len(anchor_grids)
        self.num_anchors = len(anchor_grids)
        self.strides = strides
        self.anchor_grids = anchor_grids

    def set_wh_weights(
        self,
        grid_sizes: List[List[int]],
        dtype: torch.dtype = torch.float32,
        device: torch.device = torch.device("cpu"),
    ) -> Tensor:

        wh_weights = []

        for size, stride in zip(grid_sizes, self.strides):
            grid_height, grid_width = size
            stride = torch.as_tensor([stride], dtype=dtype, device=device)
            stride = stride.view(-1, 1)
            stride = stride.repeat(1, grid_height * grid_width * self.num_anchors)
            stride = stride.reshape(-1, 1)
            wh_weights.append(stride)

        return torch.cat(wh_weights)

    def set_xy_weights(
        self,
        grid_sizes: List[List[int]],
        dtype: torch.dtype = torch.float32,
        device: torch.device = torch.device("cpu"),
    ) -> Tensor:

        xy_weights = []

        for size, anchor_grid in zip(grid_sizes, self.anchor_grids):
            grid_height, grid_width = size
            anchor_grid = torch.as_tensor(anchor_grid, dtype=dtype, device=device)
            anchor_grid = anchor_grid.view(-1, 2)
            anchor_grid = anchor_grid.repeat(1, grid_height * grid_width)
            anchor_grid = anchor_grid.reshape(-1, 2)
            xy_weights.append(anchor_grid)

        return torch.cat(xy_weights)

    def grid_anchors(
        self,
        grid_sizes: List[List[int]],
        device: torch.device = torch.device("cpu"),
    ) -> Tensor:

        anchors = []

        for size in grid_sizes:
            grid_height, grid_width = size

            # For output anchor, compute [x_center, y_center, x_center, y_center]
            shifts_x = torch.arange(0, grid_width, dtype=torch.float32, device=device)
            shifts_y = torch.arange(0, grid_height, dtype=torch.float32, device=device)
            shift_y, shift_x = torch.meshgrid(shifts_y, shifts_x)

            shifts = torch.stack((shift_x, shift_y), dim=2)
            shifts = shifts.view(1, grid_height, grid_width, 2)
            shifts = shifts.repeat(self.num_anchors, 1, 1, 1)
            shifts = shifts - torch.tensor(0.5, dtype=shifts.dtype, device=device)
            shifts = shifts.reshape(-1, 2)

            anchors.append(shifts)

        return torch.cat(anchors)

    def forward(self, feature_maps: List[Tensor]) -> Tuple[Tensor, Tensor, Tensor]:
        grid_sizes = list([feature_map.shape[-2:] for feature_map in feature_maps])
        dtype, device = feature_maps[0].dtype, feature_maps[0].device

        wh_weights = self.set_wh_weights(grid_sizes, dtype, device)
        xy_weights = self.set_xy_weights(grid_sizes, dtype, device)
        anchors = self.grid_anchors(grid_sizes, device)

        return anchors, wh_weights, xy_weights



# Modified from ultralytics/yolov5 by Zhiqiang Wang
import math

import torch
from torch import Tensor
import torch.nn.functional as F
from torchvision.ops import box_convert

from typing import Tuple


class BoxCoder(object):
    """
    This class encodes and decodes a set of bounding boxes into
    the representation used for training the regressors.
    """

    def __init__(
        self,
        weights: Tuple[float, float, float, float] = (1.0, 1.0, 1.0, 1.0),
        bbox_xform_clip: float = math.log(1000. / 16),
    ) -> None:
        """
        Arguments:
            weights (4-element tuple)
            bbox_xform_clip (float)
        """
        self.weights = weights
        self.bbox_xform_clip = bbox_xform_clip

    def decode_single(
        self,
        rel_codes: Tensor,
        anchors_tuple: Tuple[Tensor, Tensor, Tensor],
    ):
        """
        From a set of original boxes and encoded relative box offsets,
        get the decoded boxes.

        Arguments:
            rel_codes (Tensor): encoded boxes
            anchors_tupe (Tensor, Tensor, Tensor): reference boxes.
        """

        pred_wh = (rel_codes[..., 0:2] * 2. + anchors_tuple[0]) * anchors_tuple[1]  # wh
        pred_xy = (rel_codes[..., 2:4] * 2) ** 2 * anchors_tuple[2]  # xy
        pred_boxes = torch.cat([pred_wh, pred_xy], dim=1)
        pred_boxes = box_convert(pred_boxes, in_fmt="cxcywh", out_fmt="xyxy")

        return pred_boxes


def bbox_ciou(box1, box2, eps: float = 1e-9):
    # Returns the IoU of box1 to box2. box1 is 4, box2 is nx4
    box2 = box2.T

    # Get the coordinates of bounding boxes
    # transform from xywh to xyxy
    b1_x1, b1_x2 = box1[0] - box1[2] / 2, box1[0] + box1[2] / 2
    b1_y1, b1_y2 = box1[1] - box1[3] / 2, box1[1] + box1[3] / 2
    b2_x1, b2_x2 = box2[0] - box2[2] / 2, box2[0] + box2[2] / 2
    b2_y1, b2_y2 = box2[1] - box2[3] / 2, box2[1] + box2[3] / 2

    # Intersection area
    inter = (torch.min(b1_x2, b2_x2) - torch.max(b1_x1, b2_x1)).clamp(0) * \
            (torch.min(b1_y2, b2_y2) - torch.max(b1_y1, b2_y1)).clamp(0)

    # Union Area
    w1, h1 = b1_x2 - b1_x1, b1_y2 - b1_y1 + eps
    w2, h2 = b2_x2 - b2_x1, b2_y2 - b2_y1 + eps
    union = w1 * h1 + w2 * h2 - inter + eps

    iou = inter / union

    cw = torch.max(b1_x2, b2_x2) - torch.min(b1_x1, b2_x1)  # convex (smallest enclosing box) width
    ch = torch.max(b1_y2, b2_y2) - torch.min(b1_y1, b2_y1)  # convex height

    c2 = cw ** 2 + ch ** 2 + eps  # convex diagonal squared
    # center distance squared
    rho2 = ((b2_x1 + b2_x2 - b1_x1 - b1_x2) ** 2 + (b2_y1 + b2_y2 - b1_y1 - b1_y2) ** 2) / 4

    v = (4 / math.pi ** 2) * torch.pow(torch.atan(w2 / h2) - torch.atan(w1 / h1), 2)
    with torch.no_grad():
        alpha = v / ((1 + eps) - iou + v)

    return iou - (rho2 / c2 + v * alpha)  # CIoU


def cls_loss(inputs, targets, pos_weight):
    loss = F.binary_cross_entropy_with_logits(inputs, targets, pos_weight=pos_weight)
    return loss


def obj_loss(inputs, targets, pos_weight):
    loss = F.binary_cross_entropy_with_logits(inputs, targets, pos_weight=pos_weight)
    return loss



# Copyright (c) 2021, Zhiqiang Wang. All Rights Reserved.
import argparse

import torch
from torch import Tensor

import pytorch_lightning as pl

from . import yolo
from .transform import GeneralizedYOLOTransform

from typing import Any, List, Dict, Tuple, Optional


class YOLOLitWrapper(pl.LightningModule):
    """
    PyTorch Lightning wrapper of `YOLO`
    """
    def __init__(
        self,
        lr: float = 0.01,
        arch: str = 'yolov5_darknet_pan_s_r31',
        pretrained: bool = False,
        progress: bool = True,
        num_classes: int = 80,
        min_size: int = 320,
        max_size: int = 416,
        **kwargs: Any,
    ):
        """
        Args:
            arch: architecture
            lr: the learning rate
            pretrained: if true, returns a model pre-trained on COCO train2017
            num_classes: number of detection classes (including background)
        """
        super().__init__()

        self.lr = lr
        self.num_classes = num_classes

        self.model = yolo.__dict__[arch](
            pretrained=pretrained, progress=progress, num_classes=num_classes, **kwargs)

        self.transform = GeneralizedYOLOTransform(min_size, max_size)

    def forward(
        self,
        inputs: List[Tensor],
        targets: Optional[List[Dict[str, Tensor]]] = None,
    ) -> List[Dict[str, Tensor]]:
        """
        Args:
            images (list[Tensor]): images to be processed
            targets (list[Dict[Tensor]]): ground-truth boxes present in the image (optional)

        Returns:
            result (list[BoxList] or dict[Tensor]): the output from the model.
                During training, it returns a dict[Tensor] which contains the losses.
                During testing, it returns list[BoxList] contains additional fields
                like `scores`, `labels` and `mask` (for Mask R-CNN models).

        """
        # get the original image sizes
        original_image_sizes: List[Tuple[int, int]] = []
        for img in inputs:
            val = img.shape[-2:]
            assert len(val) == 2
            original_image_sizes.append((val[0], val[1]))

        # Transform the input
        samples, targets = self.transform(inputs, targets)
        # Compute the detections
        detections = self.model(samples.tensors, targets=targets)
        # Rescale coordinate
        detections = self.transform.postprocess(detections, samples.image_sizes, original_image_sizes)

        return detections

    def training_step(self, batch, batch_idx):
        """
        The training step.
        """
        # Transform the input
        samples, targets = self.transform(*batch)
        # yolov5 takes both images and targets for training, returns
        loss_dict = self.model(samples.tensors, targets)
        loss = sum(loss for loss in loss_dict.values())
        return {"loss": loss, "log": loss_dict}

    def configure_optimizers(self):
        return torch.optim.SGD(
            self.model.parameters(),
            lr=self.lr,
            momentum=0.9,
            weight_decay=0.005,
        )

    @staticmethod
    def add_model_specific_args(parent_parser):
        parser = argparse.ArgumentParser(parents=[parent_parser], add_help=False)
        parser.add_argument('--arch', default='yolov5_darknet_pan_s_r31',
                            help='model architecture')
        parser.add_argument('--num_classes', default=80, type=int,
                            help='number classes of datasets')
        parser.add_argument('--pretrained', action='store_true',
                            help='Use pre-trained models from the modelzoo')
        parser.add_argument('--lr', default=0.02, type=float,
                            help='initial learning rate, 0.02 is the default value for training '
                            'on 8 gpus and 2 images_per_gpu')
        parser.add_argument('--momentum', default=0.9, type=float, metavar='M',
                            help='momentum')
        parser.add_argument('--weight-decay', default=5e-4, type=float,
                            metavar='W', help='weight decay (default: 5e-4)')
        return parser



import unittest

import torch

from yolort.models import yolov5s, yolov5m, yolov5l


class TorchScriptTester(unittest.TestCase):
    def test_yolov5s_script(self):
        model = yolov5s(pretrained=True)
        model.eval()

        scripted_model = torch.jit.script(model)
        scripted_model.eval()

        x = [torch.rand(3, 416, 320), torch.rand(3, 480, 352)]

        out = model(x)
        out_script = scripted_model(x)
        self.assertTrue(out[0]["scores"].equal(out_script[0]["scores"]))
        self.assertTrue(out[0]["labels"].equal(out_script[0]["labels"]))
        self.assertTrue(out[0]["boxes"].equal(out_script[0]["boxes"]))

    def test_yolov5m_script(self):
        model = yolov5m(pretrained=True)
        model.eval()

        scripted_model = torch.jit.script(model)
        scripted_model.eval()

        x = [torch.rand(3, 416, 320), torch.rand(3, 480, 352)]

        out = model(x)
        out_script = scripted_model(x)
        self.assertTrue(out[0]["scores"].equal(out_script[0]["scores"]))
        self.assertTrue(out[0]["labels"].equal(out_script[0]["labels"]))
        self.assertTrue(out[0]["boxes"].equal(out_script[0]["boxes"]))

    def test_yolov5l_script(self):
        model = yolov5l(pretrained=True)
        model.eval()

        scripted_model = torch.jit.script(model)
        scripted_model.eval()

        x = [torch.rand(3, 416, 320), torch.rand(3, 480, 352)]

        out = model(x)
        out_script = scripted_model(x)
        self.assertTrue(out[0]["scores"].equal(out_script[0]["scores"]))
        self.assertTrue(out[0]["labels"].equal(out_script[0]["labels"]))
        self.assertTrue(out[0]["boxes"].equal(out_script[0]["boxes"]))


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
# Modified by Zhiqiang Wang (zhiqwang@outlook.com)
import math
import torch
from torch import nn, Tensor
import torch.nn.functional as F

import torchvision

from typing import Dict, Optional, List, Tuple


class NestedTensor(object):
    """
    Structure that holds a list of images (of possibly
    varying sizes) as a single tensor.
    This works by padding the images to the same size,
    and storing in a field the original sizes of each image
    """
    def __init__(self, tensors: Tensor, image_sizes: List[Tuple[int, int]]):
        """
        Args:
            tensors (Tensor)
            image_sizes (list[tuple[int, int]])
        """
        self.tensors = tensors
        self.image_sizes = image_sizes

    def to(self, device) -> "NestedTensor":
        cast_tensor = self.tensors.to(device)
        return NestedTensor(cast_tensor, self.image_sizes)

    def __repr__(self):
        return str(self.tensors)


class GeneralizedYOLOTransform(nn.Module):
    """
    Performs input / target transformation before feeding the data to a GeneralizedRCNN
    model.

    The transformations it perform are:
        - input normalization (mean subtraction and std division)
        - input / target resizing to match min_size / max_size

    It returns a ImageList for the inputs, and a List[Dict[Tensor]] for the targets
    """
    def __init__(self, min_size, max_size) -> None:
        super().__init__()
        if not isinstance(min_size, (list, tuple)):
            min_size = (min_size,)
        self.min_size = min_size
        self.max_size = max_size

    def forward(
        self,
        images: List[Tensor],
        targets: Optional[List[Dict[str, Tensor]]],
    ) -> Tuple[NestedTensor, Optional[Tensor]]:
        device = images[0].device
        images = [img for img in images]
        if targets is not None:
            # make a copy of targets to avoid modifying it in-place
            # once torchscript supports dict comprehension
            # this can be simplified as as follows
            # targets = [{k: v for k,v in t.items()} for t in targets]
            targets_copy: List[Dict[str, Tensor]] = []
            for t in targets:
                data: Dict[str, Tensor] = {}
                for k, v in t.items():
                    data[k] = v.to(device)
                targets_copy.append(data)
            targets = targets_copy

        for i in range(len(images)):
            image = images[i]
            target_index = targets[i] if targets is not None else None

            if image.dim() != 3:
                raise ValueError("images is expected to be a list of 3d tensors "
                                 "of shape [C, H, W], got {}".format(image.shape))

            image, target_index = self.resize(image, target_index)
            images[i] = image
            if targets is not None and target_index is not None:
                targets[i] = target_index

        image_sizes = [img.shape[-2:] for img in images]
        images = nested_tensor_from_tensor_list(images)
        image_sizes_list: List[Tuple[int, int]] = []
        for image_size in image_sizes:
            assert len(image_size) == 2
            image_sizes_list.append((image_size[0], image_size[1]))

        image_list = NestedTensor(images, image_sizes_list)

        if targets is not None:
            targets_batched = []
            for i, target in enumerate(targets):
                num_objects = len(target['labels'])
                if num_objects > 0:
                    targets_merged = torch.full((num_objects, 6), i, dtype=torch.float32, device=device)
                    targets_merged[:, 1] = target['labels']
                    targets_merged[:, 2:] = target['boxes']
                    targets_batched.append(targets_merged)
            targets_batched = torch.cat(targets_batched, dim=0)
        else:
            targets_batched = None

        return image_list, targets_batched

    def torch_choice(self, k: List[int]) -> int:
        """
        Implements `random.choice` via torch ops so it can be compiled with
        TorchScript. Remove if https://github.com/pytorch/pytorch/issues/25803
        is fixed.
        """
        index = int(torch.empty(1).uniform_(0., float(len(k))).item())
        return k[index]

    def resize(
        self,
        image: Tensor,
        target: Optional[Dict[str, Tensor]],
    ) -> Tuple[Tensor, Optional[Dict[str, Tensor]]]:

        h, w = image.shape[-2:]
        if self.training:
            size = float(self.torch_choice(self.min_size))
        else:
            # FIXME assume for now that testing uses the largest scale
            size = float(self.min_size[-1])
        if torchvision._is_tracing():
            image, target = _resize_image_and_masks_onnx(image, size, float(self.max_size), target)
        else:
            image, target = _resize_image_and_masks(image, size, float(self.max_size), target)

        if target is None:
            return image, target

        bbox = target["boxes"]
        bbox = resize_boxes(bbox, (h, w), image.shape[-2:])
        target["boxes"] = bbox

        return image, target

    def postprocess(
        self,
        result: Tuple[Dict[str, Tensor], List[Dict[str, Tensor]]],
        image_shapes: List[Tuple[int, int]],
        original_image_sizes: List[Tuple[int, int]],
    ) -> List[Dict[str, Tensor]]:

        if torch.jit.is_scripting():
            predictions = result[1]
        else:
            predictions = result

        for i, (pred, im_s, o_im_s) in enumerate(zip(predictions, image_shapes, original_image_sizes)):
            boxes = pred["boxes"]
            boxes = resize_boxes(boxes, im_s, o_im_s)
            predictions[i]["boxes"] = boxes

        return predictions


def nested_tensor_from_tensor_list(tensor_list: List[Tensor], size_divisible: int = 32):
    # TODO make this more general
    if tensor_list[0].ndim == 3:
        if torchvision._is_tracing():
            # nested_tensor_from_tensor_list() does not export well to ONNX
            # call _onnx_nested_tensor_from_tensor_list() instead
            return _onnx_nested_tensor_from_tensor_list(tensor_list, size_divisible)

        max_size = _max_by_axis([list(img.shape) for img in tensor_list])
        stride = float(size_divisible)
        max_size = list(max_size)
        max_size[1] = int(math.ceil(float(max_size[1]) / stride) * stride)
        max_size[2] = int(math.ceil(float(max_size[2]) / stride) * stride)

        batch_shape = [len(tensor_list)] + max_size
        tensor_batched = tensor_list[0].new_full(batch_shape, 0)
        for img, pad_img in zip(tensor_list, tensor_batched):
            pad_img[: img.shape[0], : img.shape[1], : img.shape[2]].copy_(img)
    else:
        raise ValueError('not supported')
    return tensor_batched


def _max_by_axis(the_list: List[List[int]]) -> List[int]:
    maxes = the_list[0]
    for sublist in the_list[1:]:
        for index, item in enumerate(sublist):
            maxes[index] = max(maxes[index], item)
    return maxes


# _onnx_nested_tensor_from_tensor_list() is an implementation of
# nested_tensor_from_tensor_list() that is supported by ONNX tracing.
@torch.jit.unused
def _onnx_nested_tensor_from_tensor_list(tensor_list: List[Tensor], size_divisible: int = 32) -> Tensor:
    max_size = []
    for i in range(tensor_list[0].dim()):
        max_size_i = torch.max(torch.stack([img.shape[i] for img in tensor_list]).to(torch.float32)).to(torch.int64)
        max_size.append(max_size_i)
    stride = size_divisible
    max_size[1] = (torch.ceil((max_size[1].to(torch.float32)) / stride) * stride).to(torch.int64)
    max_size[2] = (torch.ceil((max_size[2].to(torch.float32)) / stride) * stride).to(torch.int64)
    max_size = tuple(max_size)

    # work around for
    # pad_img[: img.shape[0], : img.shape[1], : img.shape[2]].copy_(img)
    # m[: img.shape[1], :img.shape[2]] = False
    # which is not yet supported in onnx
    padded_imgs = []

    for img in tensor_list:
        padding = [(s1 - s2) for s1, s2 in zip(max_size, tuple(img.shape))]
        padded_img = torch.nn.functional.pad(img, (0, padding[2], 0, padding[1], 0, padding[0]))
        padded_imgs.append(padded_img)

    tensor = torch.stack(padded_imgs)

    return tensor


@torch.jit.unused
def _resize_image_and_masks_onnx(
    image: Tensor,
    self_min_size: float,
    self_max_size: float,
    target: Optional[Dict[str, Tensor]],
) -> Tuple[Tensor, Optional[Dict[str, Tensor]]]:

    from torch.onnx import operators

    im_shape = operators.shape_as_tensor(image)[-2:]
    min_size = torch.min(im_shape).to(dtype=torch.float32)
    max_size = torch.max(im_shape).to(dtype=torch.float32)
    scale_factor = torch.min(self_min_size / min_size, self_max_size / max_size)

    image = torch.nn.functional.interpolate(
        image[None], scale_factor=scale_factor, mode='bilinear', recompute_scale_factor=True,
        align_corners=False)[0]

    if target is None:
        return image, target

    if "masks" in target:
        mask = target["masks"]
        mask = F.interpolate(mask[:, None].float(), scale_factor=scale_factor, recompute_scale_factor=True)[:, 0].byte()
        target["masks"] = mask
    return image, target


def _resize_image_and_masks(
    image: Tensor,
    self_min_size: float,
    self_max_size: float,
    target: Optional[Dict[str, Tensor]],
) -> Tuple[Tensor, Optional[Dict[str, Tensor]]]:

    im_shape = torch.tensor(image.shape[-2:])
    min_size = float(torch.min(im_shape))
    max_size = float(torch.max(im_shape))
    scale_factor = self_min_size / min_size
    if max_size * scale_factor > self_max_size:
        scale_factor = self_max_size / max_size
    image = torch.nn.functional.interpolate(
        image[None], scale_factor=scale_factor, mode='bilinear', recompute_scale_factor=True,
        align_corners=False)[0]

    if target is None:
        return image, target

    if "masks" in target:
        mask = target["masks"]
        mask = F.interpolate(mask[:, None].float(), scale_factor=scale_factor, recompute_scale_factor=True)[:, 0].byte()
        target["masks"] = mask
    return image, target


def resize_boxes(boxes: Tensor, original_size: List[int], new_size: List[int]) -> Tensor:
    ratios = [
        torch.tensor(s, dtype=torch.float32, device=boxes.device) /
        torch.tensor(s_orig, dtype=torch.float32, device=boxes.device)
        for s, s_orig in zip(new_size, original_size)
    ]
    ratio_height, ratio_width = ratios
    xmin, ymin, xmax, ymax = boxes.unbind(1)

    xmin = xmin * ratio_width
    xmax = xmax * ratio_width
    ymin = ymin * ratio_height
    ymax = ymax * ratio_height
    return torch.stack((xmin, ymin, xmax, ymax), dim=1)



# Copyright (c) 2021, Zhiqiang Wang. All Rights Reserved.
import torch
from torch import nn, Tensor
from torch.hub import load_state_dict_from_url

from .common import Conv, SPP, Focus, BottleneckCSP
from .experimental import C3

from typing import Callable, List, Optional, Any


__all__ = ['DarkNet', 'darknet_s_r3_1', 'darknet_m_r3_1', 'darknet_l_r3_1',
           'darknet_s_r4_0', 'darknet_m_r4_0', 'darknet_l_r4_0']

model_urls = {
    "darknet_s_r3.1": None,
    "darknet_m_r3.1": None,
    "darknet_l_r3.1": None,
    "darknet_s_r4.0": None,
    "darknet_m_r4.0": None,
    "darknet_l_r4.0": None,
}  # TODO: add checkpoint weights


def _make_divisible(v: float, divisor: int, min_value: Optional[int] = None) -> int:
    """
    This function is taken from the original tf repo.
    It ensures that all layers have a channel number that is divisible by 8
    It can be seen here:
    https://github.com/tensorflow/models/blob/master/research/slim/nets/mobilenet/mobilenet.py
    :param v:
    :param divisor:
    :param min_value:
    :return:
    """
    if min_value is None:
        min_value = divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < 0.9 * v:
        new_v += divisor
    return new_v


class DarkNet(nn.Module):
    def __init__(
        self,
        depth_multiple: float,
        width_multiple: float,
        block: Optional[Callable[..., nn.Module]] = None,
        stages_repeats: Optional[List[int]] = None,
        stages_out_channels: Optional[List[int]] = None,
        num_classes: int = 1000,
        round_nearest: int = 8,
    ) -> None:
        """
        DarkNet main class

        Args:
            num_classes (int): Number of classes
            depth_multiple (float): Depth multiplier
            width_multiple (float): Width multiplier - adjusts number of channels in each layer by this amount
            round_nearest (int): Round the number of channels in each layer to be a multiple of this number
            Set to 1 to turn off rounding
            block: Module specifying inverted residual building block for darknet
        """
        super().__init__()

        if block is None:
            block = BottleneckCSP

        input_channel = 64
        last_channel = 1024

        if stages_repeats is None:
            stages_repeats = [3, 9, 9]

        if stages_out_channels is None:
            stages_out_channels = [128, 256, 512]

        # Initial an empty features list
        layers: List[nn.Module] = []

        # building first layer
        out_channel = _make_divisible(input_channel * width_multiple, round_nearest)
        layers.append(Focus(3, out_channel, k=3))
        input_channel = out_channel

        # building CSP blocks
        for depth_gain, out_channel in zip(stages_repeats, stages_out_channels):
            depth_gain = max(round(depth_gain * depth_multiple), 1)
            out_channel = _make_divisible(out_channel * width_multiple, round_nearest)
            layers.append(Conv(input_channel, out_channel, k=3, s=2))
            layers.append(block(out_channel, out_channel, n=depth_gain))
            input_channel = out_channel

        # building last CSP blocks
        last_channel = _make_divisible(last_channel * width_multiple, round_nearest)
        layers.append(Conv(input_channel, last_channel, k=3, s=2))
        layers.append(SPP(last_channel, last_channel, k=(5, 9, 13)))

        self.features = nn.Sequential(*layers)
        self.avgpool = nn.AdaptiveAvgPool2d(1)
        self.classifier = nn.Sequential(
            nn.Linear(last_channel, last_channel),
            nn.Hardswish(inplace=True),
            nn.Dropout(p=0.2, inplace=True),
            nn.Linear(last_channel, num_classes),
        )

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                pass  # nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, nn.BatchNorm2d):
                m.eps = 1e-3
                m.momentum = 0.03
            elif isinstance(m, (nn.Hardswish, nn.LeakyReLU, nn.ReLU, nn.ReLU6)):
                m.inplace = True

    def _forward_impl(self, x: Tensor) -> Tensor:
        x = self.features(x)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)

        x = self.classifier(x)

        return x

    def forward(self, x: Tensor) -> Tensor:
        return self._forward_impl(x)


def _darknet(arch: str, pretrained: bool, progress: bool, *args: Any, **kwargs: Any) -> DarkNet:
    """
    Constructs a DarkNet architecture from
    # TODO

    """
    model = DarkNet(*args, **kwargs)

    if pretrained:
        model_url = model_urls[arch]
        if model_url is None:
            raise NotImplementedError('pretrained {} is not supported as of now'.format(arch))
        else:
            state_dict = load_state_dict_from_url(model_url, progress=progress)
            model.load_state_dict(state_dict)

    return model


_block = {
    "r3.1": BottleneckCSP,
    "r4.0": C3,
}


def darknet_s_r3_1(pretrained: bool = False, progress: bool = True, **kwargs: Any) -> DarkNet:
    """
    Constructs a DarkNet with small channels, as described in release 3.1
    # TODO

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _darknet("darknet_s_r3.1", pretrained, progress,
                    0.33, 0.5, block=_block["r3.1"], **kwargs)


def darknet_m_r3_1(pretrained: bool = False, progress: bool = True, **kwargs: Any) -> DarkNet:
    """
    Constructs a DarkNet with small channels, as described in release 3.1
    # TODO

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _darknet("darknet_m_r3.1", pretrained, progress,
                    0.67, 0.75, block=_block["r3.1"], **kwargs)


def darknet_l_r3_1(pretrained: bool = False, progress: bool = True, **kwargs: Any) -> DarkNet:
    """
    Constructs a DarkNet with small channels, as described in release 3.1
    # TODO

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _darknet("darknet_l_r3.1", pretrained, progress,
                    1.0, 1.0, block=_block["r3.1"], **kwargs)


def darknet_s_r4_0(pretrained: bool = False, progress: bool = True, **kwargs: Any) -> DarkNet:
    """
    Constructs a DarkNet with small channels, as described in release 3.1
    # TODO

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _darknet("darknet_s_r4.0", pretrained, progress,
                    0.33, 0.5, block=_block["r4.0"], **kwargs)


def darknet_m_r4_0(pretrained: bool = False, progress: bool = True, **kwargs: Any) -> DarkNet:
    """
    Constructs a DarkNet with small channels, as described in release 3.1
    # TODO

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _darknet("darknet_m_r4.0", pretrained, progress,
                    0.67, 0.75, block=_block["r4.0"], **kwargs)


def darknet_l_r4_0(pretrained: bool = False, progress: bool = True, **kwargs: Any) -> DarkNet:
    """
    Constructs a DarkNet with small channels, as described in release 3.1
    # TODO

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _darknet("darknet_l_r4.0", pretrained, progress,
                    1.0, 1.0, block=_block["r4.0"], **kwargs)



"""
Plotting a graph with bad gradient nodes marked in red.
Mostly copy-paste from https://github.com/t-vi/pytorch-tvmisc/tree/master/hacks
"""
from graphviz import Digraph


def make_graph(mod, classes_to_visit=None, classes_found=None, dot=None,
               prefix="", input_preds=None, parent_dot=None):
    preds = {}

    def find_name(i, self_input, suffix=None):
        if i == self_input:
            return suffix
        cur = i.node().s('name')
        if suffix is not None:
            cur = f'{cur}.{suffix}'
        of = next(i.node().inputs())
        return find_name(of, self_input, suffix=cur)

    gr = mod.graph
    # list(traced_model.graph.nodes())[0]
    self_input = next(gr.inputs())
    self_type = self_input.type().str().split('.')[-1]
    preds[self_input] = (set(), set())  # inps, ops

    if dot is None:
        dot = Digraph(format='svg', graph_attr={'label': self_type, 'labelloc': 't'})
        # dot.attr('node', shape='box')

    seen_inpnames = set()
    seen_edges = set()

    def add_edge(dot, n1, n2):
        if (n1, n2) not in seen_edges:
            seen_edges.add((n1, n2))
            dot.edge(n1, n2)

    def make_edges(pr, inpname, name, op, edge_dot=dot):
        if op:
            if inpname not in seen_inpnames:
                seen_inpnames.add(inpname)
                label_lines = [[]]
                line_len = 0
                for w in op:
                    if line_len >= 20:
                        label_lines.append([])
                        line_len = 0
                    label_lines[-1].append(w)
                    line_len += len(w) + 1

                edge_dot.node(
                    inpname,
                    label='\n'.join([' '.join(w) for w in label_lines]),
                    shape='box',
                    style='rounded',
                )
                for p in pr:
                    add_edge(edge_dot, p, inpname)
            add_edge(edge_dot, inpname, name)
        else:
            for p in pr:
                add_edge(edge_dot, p, name)

    for nr, i in enumerate(list(gr.inputs())[1:]):
        name = f'{prefix}inp_{i.debugName()}'
        preds[i] = {name}, set()
        dot.node(name, shape='ellipse')
        if input_preds is not None:
            pr, op = input_preds[nr]
            make_edges(pr, f'inp_{name}', name, op, edge_dot=parent_dot)

    def is_relevant_type(t):
        kind = t.kind()
        if kind == 'TensorType':
            return True
        if kind in ('ListType', 'OptionalType'):
            return is_relevant_type(t.getElementType())
        if kind == 'TupleType':
            return any([is_relevant_type(tt) for tt in t.elements()])
        return False

    for n in gr.nodes():
        only_first_ops = {'aten::expand_as'}
        rel_inp_end = 1 if n.kind() in only_first_ops else None

        relevant_inputs = [i for i in list(n.inputs())[:rel_inp_end] if is_relevant_type(i.type())]
        relevant_outputs = [o for o in n.outputs() if is_relevant_type(o.type())]

        if n.kind() == 'prim::CallMethod':
            fq_submodule_name = '.'.join([
                nc for nc in list(n.inputs())[0].type().str().split('.') if not nc.startswith('__')])
            submodule_type = list(n.inputs())[0].type().str().split('.')[-1]
            submodule_name = find_name(list(n.inputs())[0], self_input)
            name = f'{prefix}.{n.output().debugName()}'
            label = f'{prefix}{submodule_name} ({submodule_type})'

            if classes_found is not None:
                classes_found.add(fq_submodule_name)

            if ((classes_to_visit is None and (not fq_submodule_name.startswith('torch.nn')
                or fq_submodule_name.startswith('torch.nn.modules.container')))
                or (classes_to_visit is not None and (submodule_type in classes_to_visit
                    or fq_submodule_name in classes_to_visit))):

                # go into subgraph
                sub_prefix = prefix + submodule_name + '.'
                with dot.subgraph(name=f'cluster_{name}') as sub_dot:
                    sub_dot.attr(label=label)
                    submod = mod
                    for k in submodule_name.split('.'):
                        submod = getattr(submod, k)

                    make_graph(
                        submod,
                        dot=sub_dot,
                        prefix=sub_prefix,
                        input_preds=[preds[i] for i in list(n.inputs())[1:]],
                        parent_dot=dot,
                        classes_to_visit=classes_to_visit,
                        classes_found=classes_found,
                    )

                for i, o in enumerate(n.outputs()):
                    preds[o] = {sub_prefix + f'out_{i}'}, set()
            else:
                dot.node(name, label=label, shape='box')
                for i in relevant_inputs:
                    pr, op = preds[i]
                    make_edges(pr, prefix + i.debugName(), name, op)
                for o in n.outputs():
                    preds[o] = {name}, set()

        elif n.kind() == 'prim::CallFunction':
            funcname = list(n.inputs())[0].type().__repr__().split('.')[-1]
            name = prefix + '.' + n.output().debugName()
            label = funcname
            dot.node(name, label=label, shape='box')
            for i in relevant_inputs:
                pr, op = preds[i]
                make_edges(pr, prefix + i.debugName(), name, op)
            for o in n.outputs():
                preds[o] = {name}, set()

        else:
            unseen_ops = {
                'prim::ListConstruct', 'prim::TupleConstruct', 'aten::index', 'aten::size',
                'aten::slice', 'aten::unsqueeze', 'aten::squeeze', 'aten::to',
                'aten::view', 'aten::permute', 'aten::transpose', 'aten::contiguous',
                'aten::permute', 'aten::Int', 'prim::TupleUnpack', 'prim::ListUnpack',
                'aten::unbind', 'aten::select', 'aten::detach', 'aten::stack',
                'aten::reshape', 'aten::split_with_sizes', 'aten::cat', 'aten::expand',
                'aten::expand_as', 'aten::_shape_as_tensor',
            }

            # probably also partially absorbing ops. :/
            absorbing_ops = ('aten::size', 'aten::_shape_as_tensor')

            label = n.kind().split('::')[-1].rstrip('_')
            pr, op = set(), set()
            for i in relevant_inputs:
                apr, aop = preds[i]
                pr |= apr
                op |= aop
            # if pr and n.kind() not in unseen_ops:
            #     print(n.kind(), n)
            if n.kind() in absorbing_ops:
                pr, op = set(), set()
            elif len(relevant_inputs) > 0 and len(relevant_outputs) > 0 and n.kind() not in unseen_ops:
                op.add(label)
            for o in n.outputs():
                preds[o] = pr, op

    for i, o in enumerate(gr.outputs()):
        name = prefix + f'out_{i}'
        dot.node(name, shape='ellipse')
        pr, op = preds[o]
        make_edges(pr, f'inp_{name}', name, op)
    return dot



import torch
import torchvision


class ConvertVOCtoCOCO(object):

    CLASSES = (
        'aeroplane', 'bicycle', 'bird', 'boat', 'bottle',
        'bus', 'car', 'cat', 'chair', 'cow',
        'diningtable', 'dog', 'horse', 'motorbike', 'person',
        'pottedplant', 'sheep', 'sofa', 'train', 'tvmonitor',
    )

    def __call__(self, image, target):
        # return image, target
        anno = target['annotations']
        filename = anno['filename'].split('.')[0]
        image_id = target['image_id']
        image_id = torch.tensor([image_id])

        height, width = anno['size']['height'], anno['size']['width']

        boxes = []
        classes = []
        ishard = []
        objects = anno['object']
        if not isinstance(objects, list):
            objects = [objects]
        for obj in objects:
            bbox = obj['bndbox']
            bbox = [int(bbox[n]) - 1 for n in ['xmin', 'ymin', 'xmax', 'ymax']]
            boxes.append(bbox)
            classes.append(self.CLASSES.index(obj['name']))
            ishard.append(int(obj['difficult']))

        boxes = torch.as_tensor(boxes, dtype=torch.float32).reshape(-1, 4)
        classes = torch.as_tensor(classes, dtype=torch.int64)
        ishard = torch.as_tensor(ishard, dtype=torch.int64)

        target = {}
        target['boxes'] = boxes
        target['labels'] = classes
        target['ishard'] = ishard

        target['image_id'] = image_id
        target["orig_size"] = torch.as_tensor([int(height), int(width)])
        target["size"] = torch.as_tensor([int(height), int(width)])
        # convert filename in int8
        target['filename'] = torch.tensor([ord(i) for i in list(filename)], dtype=torch.int8)

        return image, target


class VOCDetection(torchvision.datasets.VOCDetection):
    def __init__(self, img_folder, year, image_set, transforms):
        super().__init__(img_folder, year=year, image_set=image_set)
        self._transforms = transforms
        self.prepare = ConvertVOCtoCOCO()

    def __getitem__(self, index):
        img, target = super().__getitem__(index)
        target = {
            'image_id': index,
            'annotations': target['annotation'],
        }
        img, target = self.prepare(img, target)
        if self._transforms is not None:
            img, target = self._transforms(img, target)

        return img, target



# -*- coding: utf-8 -*-
# Copyright (c) Facebook, Inc. and its affiliates.

"""
Note:
For your custom dataset, there is no need to hard-code metadata anywhere in the code.
For example, for COCO-format dataset, metadata will be obtained automatically
when calling `load_coco_json`. For other dataset, metadata may also be obtained in other ways
during loading.
However, we hard-coded metadata for a few common dataset here.
The only goal is to allow users who don't have these dataset to use pre-trained models.
Users don't have to download a COCO json (which contains metadata), in order to visualize a
COCO model (with correct class names and colors).
"""


COCO_CATEGORIES = [
    {"id": 1, "color": [220, 20, 60], "isthing": 1, "name": "person", "supercategory": "person"},
    {"id": 2, "color": [119, 11, 32], "isthing": 1, "name": "bicycle", "supercategory": "vehicle"},
    {"id": 3, "color": [0, 0, 142], "isthing": 1, "name": "car", "supercategory": "vehicle"},
    {"id": 4, "color": [0, 0, 230], "isthing": 1, "name": "motorcycle", "supercategory": "vehicle"},
    {"id": 5, "color": [106, 0, 228], "isthing": 1, "name": "airplane", "supercategory": "vehicle"},
    {"id": 6, "color": [0, 60, 100], "isthing": 1, "name": "bus", "supercategory": "vehicle"},
    {"id": 7, "color": [0, 80, 100], "isthing": 1, "name": "train", "supercategory": "vehicle"},
    {"id": 8, "color": [0, 0, 70], "isthing": 1, "name": "truck", "supercategory": "vehicle"},
    {"id": 9, "color": [0, 0, 192], "isthing": 1, "name": "boat", "supercategory": "vehicle"},
    {"id": 10, "color": [250, 170, 30], "isthing": 1, "name": "traffic light", "supercategory": "outdoor"},
    {"id": 11, "color": [100, 170, 30], "isthing": 1, "name": "fire hydrant", "supercategory": "outdoor"},
    {"id": 13, "color": [220, 220, 0], "isthing": 1, "name": "stop sign", "supercategory": "outdoor"},
    {"id": 14, "color": [175, 116, 175], "isthing": 1, "name": "parking meter", "supercategory": "outdoor"},
    {"id": 15, "color": [250, 0, 30], "isthing": 1, "name": "bench", "supercategory": "outdoor"},
    {"id": 16, "color": [165, 42, 42], "isthing": 1, "name": "bird", "supercategory": "animal"},
    {"id": 17, "color": [255, 77, 255], "isthing": 1, "name": "cat", "supercategory": "animal"},
    {"id": 18, "color": [0, 226, 252], "isthing": 1, "name": "dog", "supercategory": "animal"},
    {"id": 19, "color": [182, 182, 255], "isthing": 1, "name": "horse", "supercategory": "animal"},
    {"id": 20, "color": [0, 82, 0], "isthing": 1, "name": "sheep", "supercategory": "animal"},
    {"id": 21, "color": [120, 166, 157], "isthing": 1, "name": "cow", "supercategory": "animal"},
    {"id": 22, "color": [110, 76, 0], "isthing": 1, "name": "elephant", "supercategory": "animal"},
    {"id": 23, "color": [174, 57, 255], "isthing": 1, "name": "bear", "supercategory": "animal"},
    {"id": 24, "color": [199, 100, 0], "isthing": 1, "name": "zebra", "supercategory": "animal"},
    {"id": 25, "color": [72, 0, 118], "isthing": 1, "name": "giraffe", "supercategory": "animal"},
    {"id": 27, "color": [255, 179, 240], "isthing": 1, "name": "backpack", "supercategory": "accessory"},
    {"id": 28, "color": [0, 125, 92], "isthing": 1, "name": "umbrella", "supercategory": "accessory"},
    {"id": 31, "color": [209, 0, 151], "isthing": 1, "name": "handbag", "supercategory": "accessory"},
    {"id": 32, "color": [188, 208, 182], "isthing": 1, "name": "tie", "supercategory": "accessory"},
    {"id": 33, "color": [0, 220, 176], "isthing": 1, "name": "suitcase", "supercategory": "accessory"},
    {"id": 34, "color": [255, 99, 164], "isthing": 1, "name": "frisbee", "supercategory": "sports"},
    {"id": 35, "color": [92, 0, 73], "isthing": 1, "name": "skis", "supercategory": "sports"},
    {"id": 36, "color": [133, 129, 255], "isthing": 1, "name": "snowboard", "supercategory": "sports"},
    {"id": 37, "color": [78, 180, 255], "isthing": 1, "name": "sports ball", "supercategory": "sports"},
    {"id": 38, "color": [0, 228, 0], "isthing": 1, "name": "kite", "supercategory": "sports"},
    {"id": 39, "color": [174, 255, 243], "isthing": 1, "name": "baseball bat", "supercategory": "sports"},
    {"id": 40, "color": [45, 89, 255], "isthing": 1, "name": "baseball glove", "supercategory": "sports"},
    {"id": 41, "color": [134, 134, 103], "isthing": 1, "name": "skateboard", "supercategory": "sports"},
    {"id": 42, "color": [145, 148, 174], "isthing": 1, "name": "surfboard", "supercategory": "sports"},
    {"id": 43, "color": [255, 208, 186], "isthing": 1, "name": "tennis racket", "supercategory": "sports"},
    {"id": 44, "color": [197, 226, 255], "isthing": 1, "name": "bottle", "supercategory": "kitchen"},
    {"id": 46, "color": [171, 134, 1], "isthing": 1, "name": "wine glass", "supercategory": "kitchen"},
    {"id": 47, "color": [109, 63, 54], "isthing": 1, "name": "cup", "supercategory": "kitchen"},
    {"id": 48, "color": [207, 138, 255], "isthing": 1, "name": "fork", "supercategory": "kitchen"},
    {"id": 49, "color": [151, 0, 95], "isthing": 1, "name": "knife", "supercategory": "kitchen"},
    {"id": 50, "color": [9, 80, 61], "isthing": 1, "name": "spoon", "supercategory": "kitchen"},
    {"id": 51, "color": [84, 105, 51], "isthing": 1, "name": "bowl", "supercategory": "kitchen"},
    {"id": 52, "color": [74, 65, 105], "isthing": 1, "name": "banana", "supercategory": "food"},
    {"id": 53, "color": [166, 196, 102], "isthing": 1, "name": "apple", "supercategory": "food"},
    {"id": 54, "color": [208, 195, 210], "isthing": 1, "name": "sandwich", "supercategory": "food"},
    {"id": 55, "color": [255, 109, 65], "isthing": 1, "name": "orange", "supercategory": "food"},
    {"id": 56, "color": [0, 143, 149], "isthing": 1, "name": "broccoli", "supercategory": "food"},
    {"id": 57, "color": [179, 0, 194], "isthing": 1, "name": "carrot", "supercategory": "food"},
    {"id": 58, "color": [209, 99, 106], "isthing": 1, "name": "hot dog", "supercategory": "food"},
    {"id": 59, "color": [5, 121, 0], "isthing": 1, "name": "pizza", "supercategory": "food"},
    {"id": 60, "color": [227, 255, 205], "isthing": 1, "name": "donut", "supercategory": "food"},
    {"id": 61, "color": [147, 186, 208], "isthing": 1, "name": "cake", "supercategory": "food"},
    {"id": 62, "color": [153, 69, 1], "isthing": 1, "name": "chair", "supercategory": "furniture"},
    {"id": 63, "color": [3, 95, 161], "isthing": 1, "name": "couch", "supercategory": "furniture"},
    {"id": 64, "color": [163, 255, 0], "isthing": 1, "name": "potted plant", "supercategory": "furniture"},
    {"id": 65, "color": [119, 0, 170], "isthing": 1, "name": "bed", "supercategory": "furniture"},
    {"id": 67, "color": [0, 182, 199], "isthing": 1, "name": "dining table", "supercategory": "furniture"},
    {"id": 70, "color": [0, 165, 120], "isthing": 1, "name": "toilet", "supercategory": "furniture"},
    {"id": 72, "color": [183, 130, 88], "isthing": 1, "name": "tv", "supercategory": "electronic"},
    {"id": 73, "color": [95, 32, 0], "isthing": 1, "name": "laptop", "supercategory": "electronic"},
    {"id": 74, "color": [130, 114, 135], "isthing": 1, "name": "mouse", "supercategory": "electronic"},
    {"id": 75, "color": [110, 129, 133], "isthing": 1, "name": "remote", "supercategory": "electronic"},
    {"id": 76, "color": [166, 74, 118], "isthing": 1, "name": "keyboard", "supercategory": "electronic"},
    {"id": 77, "color": [219, 142, 185], "isthing": 1, "name": "cell phone", "supercategory": "electronic"},
    {"id": 78, "color": [79, 210, 114], "isthing": 1, "name": "microwave", "supercategory": "appliance"},
    {"id": 79, "color": [178, 90, 62], "isthing": 1, "name": "oven", "supercategory": "appliance"},
    {"id": 80, "color": [65, 70, 15], "isthing": 1, "name": "toaster", "supercategory": "appliance"},
    {"id": 81, "color": [127, 167, 115], "isthing": 1, "name": "sink", "supercategory": "appliance"},
    {"id": 82, "color": [59, 105, 106], "isthing": 1, "name": "refrigerator", "supercategory": "appliance"},
    {"id": 84, "color": [142, 108, 45], "isthing": 1, "name": "book", "supercategory": "indoor"},
    {"id": 85, "color": [196, 172, 0], "isthing": 1, "name": "clock", "supercategory": "indoor"},
    {"id": 86, "color": [95, 54, 80], "isthing": 1, "name": "vase", "supercategory": "indoor"},
    {"id": 87, "color": [128, 76, 255], "isthing": 1, "name": "scissors", "supercategory": "indoor"},
    {"id": 88, "color": [201, 57, 1], "isthing": 1, "name": "teddy bear", "supercategory": "indoor"},
    {"id": 89, "color": [246, 0, 122], "isthing": 1, "name": "hair drier", "supercategory": "indoor"},
    {"id": 90, "color": [191, 162, 208], "isthing": 1, "name": "toothbrush", "supercategory": "indoor"},
]



import time
from pathlib import Path

from numpy import random

import cv2
import torch

from .utils.image_utils import read_image, load_names, overlay_boxes
from .models import yolov5s


@torch.no_grad()
def inference(model, img_name, device, is_half=False):
    model.eval()
    img = cv2.imread(img_name)
    img = read_image(img, is_half=is_half)
    img = img.to(device)
    t1 = time.time()
    detections = model([img])
    time_consume = time.time() - t1

    return detections, time_consume


def main(args):
    print(args)
    device = torch.device("cuda") if torch.cuda.is_available() and args.gpu else torch.device("cpu")

    model = yolov5s(
        pretrained=True,
        min_size=args.min_size,
        max_size=args.max_size,
        score_thresh=args.score_thresh,
        nms_thresh=args.nms_thresh,
    )
    model.eval()
    model = model.to(device)

    # Initialize

    # half = device.type != 'cpu'  # half precision only supported on CUDA
    is_half = False

    # Load model
    if is_half:
        model.half()  # to FP16

    # Get names and colors
    args.names = load_names(Path(args.labelmap))
    args.colors = [[random.randint(0, 255) for _ in range(3)] for _ in range(len(args.names))]

    # Run inference
    img = torch.zeros((3, 320, 320), device=device)  # init img
    if is_half:
        img = img.half()
    _ = model([img])  # run once

    t0 = time.time()
    model_out, time_consume = inference(model, args.input_source, device, is_half)
    total_time_consume = time.time() - t0

    # Process detections
    _ = overlay_boxes(model_out, args.input_source, time_consume, args)

    if args.save_txt or args.save_img:
        print(f'Results saved to {args.output_dir}')

    print('Total time: (%.3fs)' % (total_time_consume))


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument('--labelmap', type=str, default='./notebooks/assets/coco.names',
                        help='path where the coco category in')
    parser.add_argument('--input_source', type=str, default='./test/assets/zidane.jpg',
                        help='path where the source images in')
    parser.add_argument('--output_dir', type=str, default='./data-bin/output',
                        help='path where to save')
    parser.add_argument('--min_size', type=int, default=640,
                        help='inference min size (pixels)')
    parser.add_argument('--max_size', type=int, default=640,
                        help='inference min size (pixels)')
    parser.add_argument('--score_thresh', type=float, default=0.15,
                        help='object confidence threshold')
    parser.add_argument('--nms_thresh', type=float, default=0.5,
                        help='IOU threshold for NMS')
    parser.add_argument('--gpu', action='store_true',
                        help='GPU switch')
    parser.add_argument('--save_txt', action='store_true',
                        help='save results to *.txt')
    parser.add_argument('--save_img', action='store_true',
                        help='save image inference results')
    parser.add_argument('--classes', nargs='+', type=int,
                        help='filter by class: --class 0, or --class 0 2 3')
    parser.add_argument('--agnostic_nms', action='store_true',
                        help='class-agnostic NMS')
    parser.add_argument('--augment', action='store_true',
                        help='augmented inference')

    args = parser.parse_args()

    if args.output_dir:
        Path(args.output_dir).mkdir(parents=True, exist_ok=True)

    main(args)



import unittest
import numpy as np

import torch
from torch import Tensor

from yolort.utils.image_utils import box_cxcywh_to_xyxy, letterbox, read_image, scale_coords


class ImageUtilsTester(unittest.TestCase):
    def test_read_image(self):
        N, H, W = 3, 720, 360
        img = np.random.randint(0, 255, (H, W, N), dtype='uint8')  # As a dummy image
        out = read_image(img)

        self.assertIsInstance(out, Tensor)
        self.assertEqual(tuple(out.shape), (N, H, W))


    def test_letterbox(self):
        img = np.random.randint(0, 255, (720, 360, 3), dtype='uint8')  # As a dummy image
        out = letterbox(img, new_shape=(416, 416))[0]
        self.assertEqual(tuple(out.shape), (416, 224, 3))

    def test_box_cxcywh_to_xyxy(self):
        box_cxcywh = np.asarray([[50, 50, 100, 100],
                                 [0, 0, 0, 0],
                                 [20, 25, 20, 20],
                                 [58, 65, 70, 60]], dtype=np.float32)
        exp_xyxy = np.asarray([[0, 0, 100, 100],
                               [0, 0, 0, 0],
                               [10, 15, 30, 35],
                               [23, 35, 93, 95]], dtype=np.float32)

        box_xyxy = box_cxcywh_to_xyxy(box_cxcywh)
        self.assertEqual(exp_xyxy.shape, (4, 4))
        self.assertEqual(exp_xyxy.dtype, box_xyxy.dtype)
        self.assertIsNone(np.testing.assert_array_equal(exp_xyxy, box_xyxy))

    def test_scale_coords(self):
        box_tensor = torch.tensor([[0., 0., 100., 100.],
                                   [0., 0., 0., 0.],
                                   [10., 15., 30., 35.],
                                   [20., 35., 90., 95.]], dtype=torch.float)
        exp_coords = torch.tensor([[0., 0., 108.05, 111.25],
                                   [0., 0., 0., 0.],
                                   [7.9250, 16.6875, 30.1750, 38.9375],
                                   [19.05, 38.9375, 96.9250, 105.6875]], dtype=torch.float)
        TOLERANCE = 1e-5

        box_coords_scaled = scale_coords(box_tensor, (160, 128), (178, 136))
        self.assertEqual(tuple(box_coords_scaled.shape), (4, 4))
        self.assertTrue((exp_coords - box_coords_scaled).abs().max() < TOLERANCE)

if __name__ == '__main__':
    unittest.main()



# !/usr/bin/env python
from pathlib import Path
import subprocess
from setuptools import setup, find_packages

PATH_ROOT = Path(__file__).parent
VERSION = "0.3.0rc1"

PACKAGE_NAME = 'yolort'
sha = 'Unknown'

try:
    sha = subprocess.check_output(['git', 'rev-parse', 'HEAD'], cwd=PATH_ROOT).decode('ascii').strip()
except Exception:
    pass


def write_version_file():
    version_path = PATH_ROOT / PACKAGE_NAME / 'version.py'
    with open(version_path, 'w') as f:
        f.write(f"__version__ = '{VERSION}'\n")
        f.write(f"git_version = {repr(sha)}\n")
        f.write("from torchvision.extension import _check_cuda_version\n")
        f.write("if _check_cuda_version() > 0:\n")
        f.write("    cuda = _check_cuda_version()\n")


def load_requirements(path_dir=PATH_ROOT, file_name='requirements.txt', comment_char='#'):
    with open(path_dir.joinpath(file_name), 'r', encoding="utf-8", errors="ignore") as file:
        lines = [ln.strip() for ln in file.readlines()]
    reqs = []
    for ln in lines:
        if comment_char in ln:  # filer all comments
            ln = ln[:ln.index(comment_char)].strip()
        if ln.startswith('http'):  # skip directly installed dependencies
            continue
        if ln:  # if requirement is not empty
            reqs.append(ln)
    return reqs


if __name__ == "__main__":
    print("Building wheel {}-{}".format(PACKAGE_NAME, VERSION))

    write_version_file()

    with open('README.md', encoding='utf-8') as f:
        readme = f.read()

    setup(
        name=PACKAGE_NAME,
        version=VERSION,
        description="Yet Another YOLOv5 and with Additional Runtime Stack",
        author="Zhiqiang Wang",
        author_email="me@zhiqwang.com",
        long_description=readme,
        long_description_content_type="text/markdown",
        url="https://github.com/zhiqwang/yolov5-rt-stack",
        license="GPL-3.0",
        packages=find_packages(exclude=['test', 'deployment', 'notebooks']),

        zip_safe=False,
        classifiers=[
            "Operating System :: POSIX :: Linux",
            # How mature is this project? Common values are
            #   3 - Alpha, 4 - Beta, 5 - Production/Stable
            "Development Status :: 4 - Beta",
            # Indicate who your project is intended for
            "Intended Audience :: Developers",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            "Topic :: Scientific/Engineering :: Information Analysis",
            # Pick your license as you wish
            "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
            # Specify the Python versions you support here. In particular, ensure
            # that you indicate whether you support Python 2, Python 3 or both.
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
        ],
        install_requires=load_requirements(),
    )



# Copyright (c) 2020, Zhiqiang Wang. All Rights Reserved.
import argparse
import torch

from ..models import yolov5m


def update_ultralytics_checkpoints(model, checkpoint_path_ultralytics):
    """
    It's limited that ultralytics saved model must load in their root path.
    So a very important thing is to desensitize the path befor updating
    ultralytics's trained model as following:

        >>> checkpoints_ = torch.load(weights, map_location='cpu')['model']
        >>> torch.save(checkpoints_.state_dict(), './checkpoints/yolov5s_ultralytics.pt')
    """
    state_dict = torch.load(checkpoint_path_ultralytics, map_location="cpu")

    # Update backbone features
    for name, params in model.backbone.body.named_parameters(prefix='model'):
        params.data.copy_(state_dict[name])

    for name, buffers in model.backbone.body.named_buffers(prefix='model'):
        buffers.copy_(state_dict[name])

    inner_block_maps = {'0': '9', '1': '10', '3': '13', '4': '14'}
    layer_block_maps = {'0': '17', '1': '18', '2': '20', '3': '21', '4': '23'}

    # Update PAN features
    for name, params in model.backbone.pan.inner_blocks.named_parameters():
        state_key = name.split('.')
        params.data.copy_(state_dict[f"model.{'.'.join([inner_block_maps[state_key[0]]] + state_key[1:])}"])

    for name, buffers in model.backbone.pan.inner_blocks.named_buffers():
        state_key = name.split('.')
        buffers.copy_(state_dict[f"model.{'.'.join([inner_block_maps[state_key[0]]] + state_key[1:])}"])

    for name, params in model.backbone.pan.layer_blocks.named_parameters():
        state_key = name.split('.')
        params.data.copy_(state_dict[f"model.{'.'.join([layer_block_maps[state_key[0]]] + state_key[1:])}"])

    for name, buffers in model.backbone.pan.layer_blocks.named_buffers():
        state_key = name.split('.')
        buffers.copy_(state_dict[f"model.{'.'.join([layer_block_maps[state_key[0]]] + state_key[1:])}"])

    # Update box heads
    for name, params in model.head.named_parameters(prefix='model.24'):
        params.data.copy_(state_dict[name.replace('head', 'm')])

    for name, buffers in model.head.named_buffers(prefix='model.24'):
        buffers.copy_(state_dict[name.replace('head', 'm')])

    return model


def main(args):
    model = yolov5m(pretrained=False, score_thresh=0.25)
    model = update_ultralytics_checkpoints(model, args.checkpoint_path_ultralytics)
    model = model.half()
    torch.save(model.state_dict(), args.checkpoint_path_rt_stack)


def get_args_parser():
    parser = argparse.ArgumentParser('YOLO checkpoint configures', add_help=False)
    parser.add_argument('--checkpoint_path_ultralytics', default='.checkpoints/yolov5s_ultralytics.pt',
                        help='Path of ultralytics trained yolov5 checkpoint model')
    parser.add_argument('--checkpoint_path_rt_stack', default='./checkpoints/yolov5s_rt.pt',
                        help='Path of updated yolov5 checkpoint model')

    return parser


if __name__ == "__main__":
    parser = argparse.ArgumentParser('Update checkpoint from ultralytics yolov5', parents=[get_args_parser()])
    args = parser.parse_args()

    main(args)



# Copyright (c) 2021, Zhiqiang Wang. All Rights Reserved.
import torch
from torch import nn, Tensor

from .common import Conv, BottleneckCSP

from typing import Callable, List, Dict, Optional


class PathAggregationNetwork(nn.Module):
    """
    Module that adds a PAN from on top of a set of feature maps. This is based on
    `"Path Aggregation Network for Instance Segmentation" <https://arxiv.org/abs/1803.01534>`_.

    The feature maps are currently supposed to be in increasing depth
    order.

    The input to the model is expected to be an OrderedDict[Tensor], containing
    the feature maps on top of which the PAN will be added.

    Args:
        in_channels_list (list[int]): number of channels for each feature map that
            is passed to the module
        out_channels (int): number of channels of the PAN representation

    Examples::

        >>> m = PathAggregationNetwork()
        >>> # get some dummy data
        >>> x = OrderedDict()
        >>> x['feat0'] = torch.rand(1, 128, 52, 44)
        >>> x['feat2'] = torch.rand(1, 256, 26, 22)
        >>> x['feat3'] = torch.rand(1, 512, 13, 11)
        >>> # compute the PAN on top of x
        >>> output = m(x)
        >>> print([(k, v.shape) for k, v in output.items()])
        >>> # returns
        >>>   [('feat0', torch.Size([1, 128, 52, 44])),
        >>>    ('feat2', torch.Size([1, 256, 26, 22])),
        >>>    ('feat3', torch.Size([1, 512, 13, 11]))]

    """
    def __init__(
        self,
        in_channels_list: List[int],
        depth_multiple: float,
        block: Optional[Callable[..., nn.Module]] = None,
    ):
        super().__init__()
        assert len(in_channels_list) == 3, "currently only support length 3."

        if block is None:
            block = BottleneckCSP

        depth_gain = max(round(3 * depth_multiple), 1)

        inner_blocks = [
            block(in_channels_list[2], in_channels_list[2], n=depth_gain, shortcut=False),
            Conv(in_channels_list[2], in_channels_list[1], 1, 1),
            nn.Upsample(scale_factor=2),
            block(in_channels_list[2], in_channels_list[1], n=depth_gain, shortcut=False),
            Conv(in_channels_list[1], in_channels_list[0], 1, 1),
            nn.Upsample(scale_factor=2),
        ]

        self.inner_blocks = nn.ModuleList(inner_blocks)

        layer_blocks = [
            block(in_channels_list[1], in_channels_list[0], n=depth_gain, shortcut=False),
            Conv(in_channels_list[0], in_channels_list[0], 3, 2),
            block(in_channels_list[1], in_channels_list[1], n=depth_gain, shortcut=False),
            Conv(in_channels_list[1], in_channels_list[1], 3, 2),
            block(in_channels_list[2], in_channels_list[2], n=depth_gain, shortcut=False),
        ]
        self.layer_blocks = nn.ModuleList(layer_blocks)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                pass  # nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, nn.BatchNorm2d):
                m.eps = 1e-3
                m.momentum = 0.03
            elif isinstance(m, (nn.Hardswish, nn.LeakyReLU, nn.ReLU, nn.ReLU6)):
                m.inplace = True

    def get_result_from_inner_blocks(self, x: Tensor, idx: int) -> Tensor:
        """
        This is equivalent to self.inner_blocks[idx](x),
        but torchscript doesn't support this yet
        """
        num_blocks = len(self.inner_blocks)
        if idx < 0:
            idx += num_blocks
        i = 0
        out = x
        for module in self.inner_blocks:
            if i == idx:
                out = module(x)
            i += 1
        return out

    def get_result_from_layer_blocks(self, x: Tensor, idx: int) -> Tensor:
        """
        This is equivalent to self.layer_blocks[idx](x),
        but torchscript doesn't support this yet
        """
        num_blocks = len(self.layer_blocks)
        if idx < 0:
            idx += num_blocks
        i = 0
        out = x
        for module in self.layer_blocks:
            if i == idx:
                out = module(x)
            i += 1
        return out

    def forward(self, x: Dict[str, Tensor]) -> List[Tensor]:
        """
        Computes the PAN for a set of feature maps.

        Args:
            x (OrderedDict[Tensor]): feature maps for each feature level.

        Returns:
            results (OrderedDict[Tensor]): feature maps after PAN layers.
                They are ordered from highest resolution first.
        """
        # unpack OrderedDict into two lists for easier handling
        x = list(x.values())

        # Descending the feature pyramid
        inners = []
        last_inner = self.get_result_from_inner_blocks(x[2], 0)
        last_inner = self.get_result_from_inner_blocks(last_inner, 1)
        inners.append(last_inner)
        last_inner = self.get_result_from_inner_blocks(last_inner, 2)
        last_inner = torch.cat([last_inner, x[1]], dim=1)
        last_inner = self.get_result_from_inner_blocks(last_inner, 3)
        last_inner = self.get_result_from_inner_blocks(last_inner, 4)
        inners.insert(0, last_inner)
        last_inner = self.get_result_from_inner_blocks(last_inner, 5)
        last_inner = torch.cat([last_inner, x[0]], dim=1)
        inners.insert(0, last_inner)

        # Ascending the feature pyramid
        results = []
        last_inner = self.get_result_from_layer_blocks(inners[0], 0)
        results.append(last_inner)

        for idx in range(len(inners) - 1):
            last_inner = self.get_result_from_layer_blocks(last_inner, 2 * idx + 1)
            last_inner = torch.cat([last_inner, inners[idx + 1]], dim=1)
            last_inner = self.get_result_from_layer_blocks(last_inner, 2 * idx + 2)
            results.append(last_inner)

        return results



import xml.etree.ElementTree as ET
import os
import shutil

import numpy as np


def parse_rec(filename):
    """ Parse a PASCAL VOC xml file """
    tree = ET.parse(filename)
    objects = []
    for obj in tree.findall('object'):
        obj_struct = {}
        obj_struct['name'] = obj.find('name').text
        # obj_struct['pose'] = obj.find('pose').text
        # obj_struct['truncated'] = int(obj.find('truncated').text)
        obj_struct['difficult'] = int(obj.find('difficult').text)
        bbox = obj.find('bndbox')
        obj_struct['bbox'] = [
            int(bbox.find('xmin').text),
            int(bbox.find('ymin').text),
            int(bbox.find('xmax').text),
            int(bbox.find('ymax').text),
        ]
        objects.append(obj_struct)
    return objects


def voc_ap(rec, prec, use_07_metric=False):
    """Compute VOC AP given precision and recall. If use_07_metric is true, uses
    the VOC 07 11-point method (default:False).
    """
    if use_07_metric:
        # 11 point metric
        ap = 0.0
        for t in np.arange(0.0, 1.1, 0.1):
            if np.sum(rec >= t) == 0:
                p = 0
            else:
                p = np.max(prec[rec >= t])
            ap = ap + p / 11.0
    else:
        # correct AP calculation
        # first append sentinel values at the end
        mrec = np.concatenate(([0.0], rec, [1.0]))
        mpre = np.concatenate(([0.0], prec, [0.0]))

        # compute the precision envelope
        for i in range(mpre.size - 1, 0, -1):
            mpre[i - 1] = np.maximum(mpre[i - 1], mpre[i])

        # to calculate area under PR curve, look for points
        # where X axis (recall) changes value
        i = np.where(mrec[1:] != mrec[:-1])[0]

        # and sum (\Delta recall) * prec
        ap = np.sum((mrec[i + 1] - mrec[i]) * mpre[i + 1])
    return ap


def voc_eval(detpath, annopath, imagesetfile, classname, ovthresh=0.5, use_07_metric=False):
    """
    Top level function that does the PASCAL VOC evaluation.
    detpath: Path to detections
        detpath.format(classname) should produce the detection results file.
    annopath: Path to annotations
        annopath.format(imagename) should be the xml annotations file.
    imagesetfile: Text file containing the list of images, one image per line.
    classname: Category name (duh)
    [ovthresh]: Overlap threshold (default = 0.5)
    [use_07_metric]: Whether to use VOC07's 11 point AP computation
        (default False)
    """
    # assumes detections are in detpath.format(classname)
    # assumes annotations are in annopath.format(imagename)
    # assumes imagesetfile is a text file with each line an image name

    # first load gt
    # read list of images
    with open(imagesetfile, 'r') as f:
        lines = f.readlines()
    imagenames = [x.strip() for x in lines]

    # load annots
    recs = {}
    for imagename in imagenames:
        recs[imagename] = parse_rec(annopath.format(imagename))

    # extract gt objects for this class
    class_recs = {}
    npos = 0
    for imagename in imagenames:
        R = [obj for obj in recs[imagename] if obj["name"] == classname]
        bbox = np.array([x["bbox"] for x in R])
        difficult = np.array([x["difficult"] for x in R]).astype(np.bool)
        # difficult = np.array([False for x in R]).astype(np.bool)  # treat all "difficult" as GT
        det = [False] * len(R)
        npos = npos + sum(~difficult)
        class_recs[imagename] = {"bbox": bbox, "difficult": difficult, "det": det}

    # read dets
    detfile = detpath.format(classname)
    with open(detfile, "r") as f:
        lines = f.readlines()

    splitlines = [x.strip().split(" ") for x in lines]
    image_ids = [x[0] for x in splitlines]
    confidence = np.array([float(x[1]) for x in splitlines])
    BB = np.array([[float(z) for z in x[2:]] for x in splitlines]).reshape(-1, 4)

    # sort by confidence
    sorted_ind = np.argsort(-confidence)
    BB = BB[sorted_ind, :]
    image_ids = [image_ids[x] for x in sorted_ind]

    # go down dets and mark TPs and FPs
    nd = len(image_ids)
    tp = np.zeros(nd)
    fp = np.zeros(nd)
    for d in range(nd):
        R = class_recs[image_ids[d]]
        bb = BB[d, :].astype(float)
        ovmax = -np.inf
        BBGT = R["bbox"].astype(float)

        if BBGT.size > 0:
            # compute overlaps
            # intersection
            ixmin = np.maximum(BBGT[:, 0], bb[0])
            iymin = np.maximum(BBGT[:, 1], bb[1])
            ixmax = np.minimum(BBGT[:, 2], bb[2])
            iymax = np.minimum(BBGT[:, 3], bb[3])
            iw = np.maximum(ixmax - ixmin + 1.0, 0.0)
            ih = np.maximum(iymax - iymin + 1.0, 0.0)
            inters = iw * ih

            # unions
            area_bb = (bb[2] - bb[0] + 1.0) * (bb[3] - bb[1] + 1.0)
            area_BBGT = (BBGT[:, 2] - BBGT[:, 0] + 1.0) * (BBGT[:, 3] - BBGT[:, 1] + 1.0)

            unions = area_bb + area_BBGT - inters

            overlaps = inters / unions
            ovmax = np.max(overlaps)
            jmax = np.argmax(overlaps)

        if ovmax > ovthresh:
            if not R["difficult"][jmax]:
                if not R["det"][jmax]:
                    tp[d] = 1.0
                    R["det"][jmax] = 1
                else:
                    fp[d] = 1.0
        else:
            fp[d] = 1.0

    # compute precision recall
    fp = np.cumsum(fp)
    tp = np.cumsum(tp)
    rec = tp / float(npos)
    # avoid divide by zero in case the first detection matches a difficult
    # ground truth
    prec = tp / np.maximum(tp + fp, np.finfo(np.float64).eps)
    ap = voc_ap(rec, prec, use_07_metric)

    return rec, prec, ap


def _write_voc_results_file(all_boxes, image_index, cls_names, output_dir):
    output_path = os.path.join(output_dir, 'results')
    if os.path.exists(output_path):
        shutil.rmtree(output_path)
    os.makedirs(output_path)
    print('Writing results file, wait a moment...')
    for cls_ind, cls_name in enumerate(cls_names):
        # DistributeSampler happens to clone the inputs to make the task
        # lenghts even among the nodes:
        # https://github.com/pytorch/pytorch/issues/22584
        # Boxes can be duplicated in the process since multiple
        # evaluation of the same image can happen, multiple boxes in the
        # same location decrease the final mAP, later in the code we discard
        # repeated image_index thanks to the sorting
        new_image_index, all_boxes[cls_ind] = zip(
            *sorted(zip(image_index, all_boxes[cls_ind]), key=lambda x: x[0]),
        )
        if cls_name == '__background__':
            continue

        filename = os.path.join(output_path, 'det_test_{:s}.txt'.format(cls_name))
        with open(filename, 'wt') as f:
            prev_index = ''
            for im_ind, index in enumerate(new_image_index):
                # check for repeated input and discard
                if prev_index == index:
                    continue
                prev_index = index
                dets = all_boxes[cls_ind][im_ind]

                if dets == []:
                    continue

                # the VOCdevkit expects 1-based indices
                for k in range(len(dets)):
                    f.write('{:s} {:.3f} {:.1f} {:.1f} {:.1f} {:.1f}\n'.format(
                        index,
                        dets[k][-1],
                        dets[k][0] + 1,
                        dets[k][1] + 1,
                        dets[k][2] + 1,
                        dets[k][3] + 1,
                    ))


def _do_python_eval(data_loader, output_dir, use_07=True):
    voc_root = os.path.join(data_loader.dataset.root, 'VOCdevkit', 'VOC2007')
    image_set = data_loader.dataset.image_set
    imagesetfile = os.path.join(voc_root, 'ImageSets', 'Main', '{}.txt'.format(image_set))
    annopath = os.path.join(voc_root, 'Annotations/{:s}.xml')

    cls_names = data_loader.dataset.prepare.CLASSES
    aps = []
    # The PASCAL VOC metric changed in 2010
    use_07_metric = use_07
    print('VOC07 metric? ' + ('Yes' if use_07_metric else 'No'))

    for cls_name in cls_names:
        if cls_name == '__background__':
            continue
        detpath = os.path.join(output_dir, 'results', 'det_test_{:s}.txt')
        rec, prec, ap = voc_eval(
            detpath, annopath, imagesetfile, cls_name,
            ovthresh=0.5, use_07_metric=use_07_metric,
        )
        aps += [ap]
        print('AP for {} = {:.4f}'.format(cls_name, ap))

    print('Mean AP = {:.4f}'.format(np.mean(aps)))



import unittest
import torch

from yolort.models.backbone_utils import darknet_pan_backbone
from yolort.models.anchor_utils import AnchorGenerator
from yolort.models.box_head import YoloHead, PostProcess, SetCriterion

from .common_utils import TestCase

from typing import Dict


# If 'unwrapper' is provided it will be called with the script model outputs
# before they are compared to the eager model outputs. This is useful if the
# model outputs are different between TorchScript / Eager mode
script_model_unwrapper = {
    "PostProcess": lambda x: x[1],
}


class ModelTester(TestCase):
    def _get_strides(self):
        return [8, 16, 32]

    def _get_in_channels(self):
        return [128, 256, 512]

    def _get_anchor_grids(self):
        return [
            [ 10,  13,  16,  30,  33,  23],
            [ 30,  61,  62,  45,  59, 119],
            [116,  90, 156, 198, 373, 326],
        ]

    def _get_num_classes(self):
        return 80

    def _get_num_outputs(self):
        return self._get_num_classes() + 5

    def _get_num_anchors(self):
        return len(self._get_anchor_grids())

    def _get_anchors_shape(self):
        return [(9009, 2), (9009, 1), (9009, 2)]

    def _get_feature_shapes(self, h, w):
        strides = self._get_strides()
        in_channels = self._get_in_channels()

        return [(c, h // s, w // s) for (c, s) in zip(in_channels, strides)]

    def _get_feature_maps(self, batch_size, h, w):
        feature_shapes = self._get_feature_shapes(h, w)
        feature_maps = [torch.rand(batch_size, *f_shape) for f_shape in feature_shapes]
        return feature_maps

    def _get_head_outputs(self, batch_size, h, w):
        feature_shapes = self._get_feature_shapes(h, w)

        num_anchors = self._get_num_anchors()
        num_outputs = self._get_num_outputs()
        head_shapes = [(batch_size, num_anchors, *f_shape[1:], num_outputs) for f_shape in feature_shapes]
        head_outputs = [torch.rand(*h_shape) for h_shape in head_shapes]

        return head_outputs

    def _init_test_backbone_with_fpn(self):
        backbone_name = 'darknet_s_r3_1'
        depth_multiple = 0.33
        width_multiple = 0.5
        backbone_with_fpn = darknet_pan_backbone(backbone_name, depth_multiple, width_multiple)
        return backbone_with_fpn

    def test_backbone_with_fpn(self):
        N, H, W = 4, 416, 352
        out_shape = self._get_feature_shapes(H, W)

        x = torch.rand(N, 3, H, W)
        model = self._init_test_backbone_with_fpn()
        out = model(x)

        self.assertEqual(len(out), 3)
        self.assertEqual(tuple(out[0].shape), (N, *out_shape[0]))
        self.assertEqual(tuple(out[1].shape), (N, *out_shape[1]))
        self.assertEqual(tuple(out[2].shape), (N, *out_shape[2]))
        self.check_jit_scriptable(model, (x,))

    def _init_test_anchor_generator(self):
        strides = self._get_strides()
        anchor_grids = self._get_anchor_grids()
        anchor_generator = AnchorGenerator(strides, anchor_grids)
        return anchor_generator

    def test_anchor_generator(self):
        N, H, W = 4, 416, 352
        feature_maps = self._get_feature_maps(N, H, W)
        model = self._init_test_anchor_generator()
        anchors = model(feature_maps)

        self.assertEqual(len(anchors), 3)
        self.assertEqual(tuple(anchors[0].shape), (9009, 2))
        self.assertEqual(tuple(anchors[1].shape), (9009, 1))
        self.assertEqual(tuple(anchors[2].shape), (9009, 2))
        self.check_jit_scriptable(model, (feature_maps,))

    def _init_test_yolo_head(self):
        in_channels = self._get_in_channels()
        num_anchors = self._get_num_anchors()
        num_classes = self._get_num_classes()
        box_head = YoloHead(in_channels, num_anchors, num_classes)
        return box_head

    def test_yolo_head(self):
        N, H, W = 4, 416, 352
        feature_maps = self._get_feature_maps(N, H, W)
        model = self._init_test_yolo_head()
        head_outputs = model(feature_maps)
        self.assertEqual(len(head_outputs), 3)

        target_head_outputs = self._get_head_outputs(N, H, W)

        self.assertEqual(head_outputs[0].shape, target_head_outputs[0].shape)
        self.assertEqual(head_outputs[1].shape, target_head_outputs[1].shape)
        self.assertEqual(head_outputs[2].shape, target_head_outputs[2].shape)
        self.check_jit_scriptable(model, (feature_maps,))

    def _init_test_postprocessors(self):
        score_thresh = 0.5
        nms_thresh = 0.45
        detections_per_img = 100
        postprocessors = PostProcess(score_thresh, nms_thresh, detections_per_img)
        return postprocessors

    def test_postprocessors(self):
        N, H, W = 4, 416, 352
        feature_maps = self._get_feature_maps(N, H, W)
        head_outputs = self._get_head_outputs(N, H, W)

        anchor_generator = self._init_test_anchor_generator()
        anchors_tuple = anchor_generator(feature_maps)
        model = self._init_test_postprocessors()
        out = model(head_outputs, anchors_tuple)

        self.assertEqual(len(out), N)
        self.assertIsInstance(out[0], Dict)
        self.assertIsInstance(out[0]["boxes"], torch.Tensor)
        self.assertIsInstance(out[0]["labels"], torch.Tensor)
        self.assertIsInstance(out[0]["scores"], torch.Tensor)
        self.check_jit_scriptable(model, (head_outputs, anchors_tuple))

    def _init_test_criterion(self):
        weights = (1.0, 1.0, 1.0, 1.0)
        fg_iou_thresh = 0.5
        bg_iou_thresh = 0.4
        allow_low_quality_matches = True
        criterion = SetCriterion(weights, fg_iou_thresh, bg_iou_thresh, allow_low_quality_matches)
        return criterion

    @unittest.skip("Current it isn't well implemented")
    def test_criterion(self):
        model = self._init_test_criterion()
        scripted_model = torch.jit.script(model)  # noqa


class AnchorGeneratorTester(TestCase):
    def _init_test_anchor_generator(self):
        strides = [4]
        anchor_grids = [[6, 14]]
        anchor_generator = AnchorGenerator(strides, anchor_grids)

        return anchor_generator

    def get_features(self, images):
        s0, s1 = images.shape[-2:]
        features = [torch.rand(2, 8, s0 // 5, s1 // 5)]
        return features

    def test_anchor_generator(self):
        images = torch.randn(2, 3, 10, 10)
        features = self.get_features(images)

        model = self._init_test_anchor_generator()
        model.eval()
        anchors = model(features)

        anchor_output = torch.tensor([[-0.5, -0.5], [0.5, -0.5], [-0.5, 0.5], [0.5, 0.5]])
        wh_output = torch.tensor([[4.], [4.], [4.], [4.]])
        xy_output = torch.tensor([[6., 14.], [6., 14.], [6., 14.], [6., 14.]])

        self.assertEqual(len(anchors), 3)
        self.assertEqual(tuple(anchors[0].shape), (4, 2))
        self.assertEqual(tuple(anchors[1].shape), (4, 1))
        self.assertEqual(tuple(anchors[2].shape), (4, 2))
        self.assertEqual(anchors[0], anchor_output)
        self.assertEqual(anchors[1], wh_output)
        self.assertEqual(anchors[2], xy_output)


if __name__ == '__main__':
    unittest.main()
